/* gcomposer: a little tool to write music.
 * http://sed.free.fr/gcomposer
 * Released in the public domain (freedom has no license).
 */

action delete_chord(doc *d, int line_number, int chord_number)
{
  data {
    int line_number;
    int chord_number;
    chord ch;
  }

  do {
    chord *c = &d->data[line_number].u.music.data[chord_number];
    data->line_number = line_number;
    data->chord_number = chord_number;
    data->ch = *c;

    /* we don't want to free mem, so let's force the chord to "forget"
     * it has some memory attached to it before to delete it
     */
    if (c->type == NOTES_CHORD) {
      c->u.chord.data = NULL;
      c->u.chord.size = 0;
      c->u.chord.maxsize = 0;
    }

    if (!doc_delete_chord(d, line_number, chord_number)) return 0;

    if (REGISTER_HISTORY < 0) {
      set_error(OOM);
      /* put the chord back in place */
      doc_insert_chord(d, line_number, chord_number, data->ch.type);
      c = &d->data[line_number].u.music.data[chord_number];
      *c = data->ch;
      return 0;
    }

    d->state = STATE_MODIFIED;
    size_delete_chord(d, line_number, chord_number);

    return 1;
  }

  undo {
    chord *c;
    /* put the chord back in place */
    doc_insert_chord(d, data->line_number, data->chord_number, data->ch.type);
    c = &d->data[data->line_number].u.music.data[data->chord_number];
    *c = data->ch;
    /* "forget" the saved data for not destroy it when requested */
    if (data->ch.type == NOTES_CHORD) {
      data->ch.u.chord.data = NULL;
      data->ch.u.chord.size = 0;
      data->ch.u.chord.maxsize = 0;
    }
    d->cursor.line = data->line_number;
    d->cursor.chord = data->chord_number;
    size_new_chord(d, data->line_number, data->chord_number);
    d->state = STATE_MODIFIED;
  }

  redo {
    chord *c = &d->data[data->line_number].u.music.data[data->chord_number];
    /* "remember" the chord again */
    data->ch = *c;
    if (c->type == NOTES_CHORD) {
      c->u.chord.data = NULL;
      c->u.chord.size = 0;
      c->u.chord.maxsize = 0;
    }
    doc_delete_chord(d, data->line_number, data->chord_number);
    d->cursor.line = data->line_number;
    d->cursor.chord = data->chord_number;
    size_delete_chord(d, data->line_number, data->chord_number);
    d->state = STATE_MODIFIED;
  }

  save {
  }

  load {
  }

  erase {
    destroy_chord(&data->ch);
  }
}
