/* gcomposer: a little tool to write music.
 * http://sed.free.fr/gcomposer
 * Released in the public domain (freedom has no license).
 */

action insert_note(doc *d, int line, int ch, elem_type type, ...)
{
  data {
    int line;
    int chord;
    int note;
    elem e;
  }

  do {
    int ret;
    va_list va;

    data->line = line;
    data->chord = ch;

    va_start(va, type);

    switch (type) {
      case NOTE: {
        int a = va_arg(va, int);
        note_alteration b = va_arg(va, note_alteration);
        note_duration c = va_arg(va, note_duration);
        int z = va_arg(va, int);
        int d1 = va_arg(va, int);
        int d2 = va_arg(va, int);
        int d3 = va_arg(va, int);
        int stem_down = va_arg(va, int);
        ret = doc_insert_note(d, line, ch, NOTE, a, b, c, z, d1, d2, d3,
                              stem_down);
        break;
      }
      case SILENCE: {
        int a = va_arg(va, int);
        note_duration b = va_arg(va, note_duration);
        int c = va_arg(va, int);
        ret = doc_insert_note(d, line, ch, SILENCE, a, b, c);
        break;
      }
    }

    va_end(va);

    if (!ret) return ret;

    data->note = ret - 1;

    data->e = d->data[line].u.music.data[ch].u.chord.data[data->note];

    if (REGISTER_HISTORY < 0) {
      set_error(OOM);
      doc_delete_note(d, line, ch, data->note);
      return 0;
    }

    size_change_chord(d, data->line, data->chord);
    d->state = STATE_MODIFIED;

    /* we have to return inserted note + 1 */
    return ret;
  }

  undo {
    doc_delete_note(d, data->line, data->chord, data->note);
    d->cursor.line = data->line;
    d->cursor.chord = data->chord;
    size_change_chord(d, data->line, data->chord);
    d->state = STATE_MODIFIED;
  }

  redo {
    switch (data->e.type) {
      case NOTE:
        doc_insert_note(d, data->line, data->chord, NOTE,
                        data->e.u.note.pitch,
                        data->e.u.note.alteration,
                        data->e.u.note.duration,
                        data->e.u.note.nb_dots,
                        data->e.u.note.head_delta,
                        data->e.u.note.alteration_delta,
                        data->e.u.note.dots_delta,
                        data->e.u.note.stem_down);
        break;
      case SILENCE:
        doc_insert_note(d, data->line, data->chord, SILENCE,
                        data->e.u.silence.position,
                        data->e.u.silence.duration,
                        data->e.u.silence.nb_dots);
        break;
    }
    d->cursor.line = data->line;
    d->cursor.chord = data->chord;
    size_change_chord(d, data->line, data->chord);
    d->state = STATE_MODIFIED;
  }

  save {
  }

  load {
  }

  erase {
  }
}
