/* gcomposer: a little tool to write music.
 * http://sed.free.fr/gcomposer
 * Released in the public domain (freedom has no license).
 */

action split_line(doc *d, int line_number, int before_chord)
{
  data {
    int line_number;
    int before_chord;
    void *data;
    int maxsize;
  }

  do {
    data->line_number = line_number;
    data->before_chord = before_chord;

    size_presplit_line(d, line_number);

    if (!doc_split_line(d, line_number, before_chord, NULL, 0)) return 0;

    data->data = NULL;
    data->maxsize = 0;

    if (REGISTER_HISTORY < 0) {
      set_error(OOM);
      doc_merge_lines(d, line_number, MERGE_DELETE);
      return 0;
    }

    size_split_line(d, data->line_number);
    d->state = STATE_MODIFIED;

    return 1;
  }

  undo {
    /* "remember" the soon deleted line */
    data->data = d->data[data->line_number+1].u.music.data;
    data->maxsize = d->data[data->line_number+1].u.music.maxsize;
    size_premerge_line(d, data->line_number+1);
    doc_merge_lines(d, data->line_number, MERGE_DONT_DELETE);
    d->cursor.line = data->line_number;
    d->cursor.chord = data->before_chord;
    size_merge_line(d, data->line_number);
    d->state = STATE_MODIFIED;
  }

  redo {
    size_presplit_line(d, data->line_number);
    doc_split_line(d, data->line_number, data->before_chord,
                   data->data, data->maxsize);
    /* the data is now in use by the app., "forget" it */
    data->data = NULL;
    data->maxsize = 0;
    d->cursor.line = data->line_number + 1;
    d->cursor.chord = 0;
    size_split_line(d, data->line_number);
    d->state = STATE_MODIFIED;
  }

  save {
  }

  load {
  }

  erase {
    if (data->maxsize) {
      free(data->data);
      data->data = NULL;
      data->maxsize = 0;
    }
  }
}
