/* gcomposer: a little tool to write music.
 * http://sed.free.fr/gcomposer
 * Released in the public domain (freedom has no license).
 */

#include "cursor.h"

#define CURSOR_BLINK_TIME_ON    400
#define CURSOR_BLINK_TIME_OFF   200

#include "gtk_wrapper.h"
#include "doc.h"
#include "draw.h"
#include "dim.h"
#include "size.h"
#include "dprintf.h"

#include <stdlib.h>

int cursor_callback(cursor *c)
{
  /* normally we cannot come here with hidden_level>0, but just to be sure */
  if (c->hidden_level) return 0;

  gtk_wrapper_source_remove(c->timeout_id);
  c->state = 1 - c->state;
  c->timeout_id = gtk_wrapper_timeout_add(c->state ? CURSOR_BLINK_TIME_ON
                                                   : CURSOR_BLINK_TIME_OFF,
                                          (void *)cursor_callback,
                                  c);
//  invalidate_line((doc *)c->doc, c->line);
  if (c->state) draw_cursor(c); else clear_cursor(c);
  return 0;
}

static void reset_cursor_and_maybe_show(cursor *c, int show)
{
  if (c->timeout_id) gtk_wrapper_source_remove(c->timeout_id);
  /* is the cursor disabled? */
  if (c->hidden_level) {
    c->timeout_id = 0;
//    invalidate_line((doc *)c->doc, c->line);
    clear_cursor(c);
    if (show) show_cursor(c);
    return;
  }
  c->timeout_id = gtk_wrapper_timeout_add(CURSOR_BLINK_TIME_ON,
                                          (void *)cursor_callback,
                                          c);
  c->state = 1;
//  invalidate_line((doc *)c->doc, c->line);
  draw_cursor(c);
  if (show) {
#if 0
    update_scrollbars(c->doc, 1);
#endif
    dprintf("show_cursor called from reset_cursor\n");
    show_cursor(c);
  }
}

void reset_cursor(cursor *c)
{
  dprintf("reset_cursor\n");
  dprintf("  (hidden_level = %d)\n", c->hidden_level);
  reset_cursor_and_maybe_show(c, 1);
}

void close_cursor(cursor *c)
{
  if (c->timeout_id) gtk_wrapper_source_remove(c->timeout_id);
  c->timeout_id = 0;
}

void disable_cursor(cursor *c)
{
  dprintf("disable_cursor\n");
  c->hidden_level++;
  if (c->hidden_level == 1) reset_cursor_and_maybe_show(c, 0);
}

void enable_cursor(cursor *c)
{
  dprintf("enable_cursor\n");
  c->hidden_level--;
  if (c->hidden_level == 0) {
    c->state = 1;
    reset_cursor_and_maybe_show(c, 0);
  }
}

/* position the cursor near the (x,y) position of the drawing area */
void position_cursor(cursor *c, int x, int y)
{
  int first_line;
  int last_line;
  int first_chord;
  int last_chord;
  doc *d = c->doc;
  int line;
  int chord;

  /* no need to call ensure_height_in_view because position_cursor
   * is only called when the user clicks the drawing_area, so the
   * height is necessarily valid (if the code changes, maybe it
   * won't be true anymore and we then might need to call
   * ensure_height_in_view)
   */

  /* cursor becomes big */
  c->type = BIG_CURSOR;

  /* find closest line */
  get_crossed_lines(d, y, y+1, &first_line, &last_line);
  if (first_line == -1) {
    if (y < UPPER_MARGIN)
      line = 0;
    else
      line = d->size-1;
  } else
    line = first_line;

  /* find closest chord */
  switch (LINE_TYPE(d, line)) {
    case LINE_TEXT:
      x -= LEFT_MARGIN;
      /* head or tail? */
      if (abs(x-0) < abs(x-(d->data[line].dim.width-1)))
        chord = 0;
      else
        chord = 1;
      break;
    case LINE_MUSIC:
      get_crossed_chords(d->data+line, x, x+1, &first_chord, &last_chord);
      x -= LEFT_MARGIN;
      if (first_chord == -1) {
        if (LINE_MUSIC_SIZE(d, line) &&
            x < d->data[line].u.music.data[0].dim.x)
          chord = 0;
        else
          chord = LINE_MUSIC_SIZE(d, line);
      } else {
        /* if we are closer to end of chord, then cursor is at next chord */
        if (abs(x-d->data[line].u.music.data[first_chord].dim.x) <
            abs(x-(d->data[line].u.music.data[first_chord].dim.x+
                   d->data[line].u.music.data[first_chord].dim.width-1)))
          chord = first_chord;
        else
          chord = first_chord + 1;
      }
      break;
  }

  c->line = line;
  c->chord = chord;
}
