/*
 * l: a language for cgscore to describe actions
 * public domain, like cgscore
 */

#include "actions.h"
#include "debug.h"

#include <stdio.h>
#include <string.h>

#include "parser.h"
#include "list.h"
#include "string.h"

/* barbarian external globals (reminds me my daily life of corporate coder) */
extern list include;
extern list cinclude;
extern list hinclude;
extern list actions;
extern list registered_actions;

/* 0 ok, -1 error */
int get_registered_actions(void)
{
  if (parse_registered_actions() == -1) {
    fprintf(stderr, "hell, you should not read this!\n");
    return -1;
  }

  return 0;
}

/* 0 ok, -1 error */
int register_action(int pos)
{
  int i;
  char *action_name = get_string((string)list_get_int(actions, pos));

  /* look for it in already known actions */
  SLOW; /* yeah, I should use hashtable or something, laziness... */

  for (i=0; i<list_size(registered_actions); i+=2) {
    if (strcmp(get_string((string)list_get_int(registered_actions, i)),
               action_name))
      continue;

    /* should not have been encountered in the parsing before */
    if (list_get_int(registered_actions, i+1) != -1) {
      fprintf(stderr, "action '%s' defined more than once\n", action_name);
      return -1;
    }

    list_set_int(registered_actions, i+1, pos);

    return 0;
  }

  /* not defined yet, let's define it */
  if (list_add_int(registered_actions, list_get_int(actions, pos)) == -1 ||
      list_add_int(registered_actions, pos) == -1) return -1;

  fprintf(stderr, "l: new action: %s\n", action_name);

  return 0;
}

#define N "actions.list"
#define P(x) do \
  { if (fprintf(f, "%s", x) == -1) { perror(N); return -1; }} while(0)

/* 0 ok, -1 error */
int store_actions(void)
{
  int i;
  FILE *f = fopen(N, "w");

  if (!f) { perror(N); return -1; }

  P("# This file was generated. Don't change it.\n");
  P("# The order matters a lot. If you change it, it won't be compatible\n");
  P("# anymore with the rest of the world. ");
  P("So don't change a shit in there.\n");

  for (i = 0; i<list_size(registered_actions); i+=2) {
    P(get_string((string)list_get_int(registered_actions, i)));
    P("\n");
  }

  if (fclose(f) == -1) { perror(N); return -1; }

  return 0;
}

#define C "actions.c"
char *H = "actions.h";
#define PC(x) do \
  { if (fprintf(c, "%s", x) == -1) {perror(C);return -1;}} while(0)
#define PH(x) do \
  { if (fprintf(h, "%s", x) == -1) {perror(H);return -1;}} while(0)

#define ENUM_PREFIX "ACTION_"

/* ugly globals - I work(ed) for a corporate company you know... */
static FILE *c, *h;

#undef N
char *N = "history_enum.h";

/* 0 ok, -1 error */
int put_enum(void)
{
  int i;
  FILE *f = fopen(N, "w");

  if (!f) { perror(N); return -1; }

  P("/* This file was generated by 'l'. ");
  P("You probably should not modify it. */\n");
  P("\n");
  P("#ifndef _HISTORY_ENUM_H_\n");
  P("#define _HISTORY_ENUM_H_\n");
  P("\n");
  P("#include \"typedefs.h\"\n\n");
  P("typedef enum {\n");

  for (i=0; i<list_size(registered_actions); i+=2) {
    P("  ");
    P(ENUM_PREFIX);
    P(get_string((string)list_get_int(registered_actions, i)));
    if (i != list_size(registered_actions)-2) P(",");
    P("\n");
  }

  P("} history_type;\n\n");

  P("typedef struct {\n");
  P(" void (*undo)(struct doc *, void *);\n");
  P(" void (*redo)(struct doc *, void *);\n");
  P(" void (*erase)(struct doc *, void *);\n");
  P("} history_functions;\n");
  P("\n");
  P("extern history_functions histories[];\n");
  P("\n");
  P("#define UNDO(d, i, p) histories[i].undo(d, (void*)(p))\n");
  P("#define REDO(d, i, p) histories[i].redo(d, (void*)(p))\n");
  P("#define ERASE(d, i, p) histories[i].erase(d, (void*)(p))\n");
  P("\n");

  P("#endif /* _HISTORY_ENUM_H_ */\n");

  if (fclose(f) == -1) { perror(N); return -1; }

  return 0;
}

/* 0 ok, -1 error */
int put_history(void)
{
  int i;

  /* define undo/redo functions */
  for (i=0; i<list_size(actions); i+=9) {
    PC("extern void ");
    PC(get_string((string)list_get_int(actions, i)));
    PC("_undo(doc *, void *);\n");

    PC("extern void ");
    PC(get_string((string)list_get_int(actions, i)));
    PC("_redo(doc *, void *);\n");

    PC("extern void ");
    PC(get_string((string)list_get_int(actions, i)));
    PC("_erase(doc *, void *);\n");

    PC("\n");
  }

  /* define histories array */
  PC("history_functions histories[] = {\n");
  for (i=0; i<list_size(registered_actions); i+=2) {
    PC("  { undo: ");
    PC(get_string((string)list_get_int(registered_actions, i)));
    PC("_undo, redo: ");
    PC(get_string((string)list_get_int(registered_actions, i)));
    PC("_redo, erase: ");
    PC(get_string((string)list_get_int(registered_actions, i)));
    PC("_erase },\n");
  }
  PC("};\n");

  return 0;
}

/* 0 ok, -1 error */
int put_proto(void)
{
  int i;

  for (i=0; i<list_size(actions); i+=9) {
    PH("int ");
    PH(get_string((string)list_get_int(actions, i)));
    PH("(");
    PH(get_string((string)list_get_int(actions, i+1)));
    PH(");\n");

    PH("void ");
    PH(get_string((string)list_get_int(actions, i)));
    PH("_undo(doc *, void *);\n");

    PH("void ");
    PH(get_string((string)list_get_int(actions, i)));
    PH("_redo(doc *, void *);\n");

    PH("void ");
    PH(get_string((string)list_get_int(actions, i)));
    PH("_erase(doc *, void *);\n");

    PH("\n");
  }

  return 0;
}

/* 0 ok, -1 error */
int put_undo(void)
{
  int i;

  for (i=0; i<list_size(actions); i+=9) {
    PC("void ");
    PC(get_string((string)list_get_int(actions, i)));
    PC("_undo(doc *d, void *_data)\n{\n");
    PC("  struct { ");
    PC(get_string((string)list_get_int(actions, i+2)));
    PC("} __data, *data = &__data;\n");
    PC("  memcpy(data, _data, sizeof(__data));\n");
    PC("  ");
    PC(get_string((string)list_get_int(actions, i+4)));
    PC("\n  memcpy(_data, data, sizeof(__data));\n}\n\n");
  }

  return 0;
}

/* 0 ok, -1 error */
int put_redo(void)
{
  int i;

  for (i=0; i<list_size(actions); i+=9) {
    PC("void ");
    PC(get_string((string)list_get_int(actions, i)));
    PC("_redo(doc *d, void *_data)\n{\n");
    PC("  struct { ");
    PC(get_string((string)list_get_int(actions, i+2)));
    PC("} __data, *data = &__data;\n");
    PC("  memcpy(data, _data, sizeof(__data));\n");
    PC("  ");
    PC(get_string((string)list_get_int(actions, i+5)));
    PC("\n  memcpy(_data, data, sizeof(__data));\n}\n\n");
  }

  return 0;
}

/* 0 ok, -1 error */
int put_erase(void)
{
  int i;

  for (i=0; i<list_size(actions); i+=9) {
    PC("void ");
    PC(get_string((string)list_get_int(actions, i)));
    PC("_erase(doc *d, void *_data)\n{\n");
    PC("  struct { ");
    PC(get_string((string)list_get_int(actions, i+2)));
    PC("} __data, *data = &__data;\n");
    PC("  memcpy(data, _data, sizeof(__data));\n");
    PC("  ");
    PC(get_string((string)list_get_int(actions, i+8)));
    PC("\n  memcpy(_data, data, sizeof(__data));\n}\n\n");
  }

  return 0;
}

/* 0 ok, -1 error */
int put_function(void)
{
  int i;

  for (i=0; i<list_size(actions); i+=9) {
    PC("#define REGISTER_HISTORY add_history(d, ACTION_");
    PC(get_string((string)list_get_int(actions, i)));
    PC(", data, sizeof(_data))\n");
    PC("int ");
    PC(get_string((string)list_get_int(actions, i)));
    PC("(");
    if (list_get_int(actions, i+1) != -1)
      PC(get_string((string)list_get_int(actions, i+1)));
    else
      PC("void");
    PC(")\n{\n");
    PC("  struct { ");
    PC(get_string((string)list_get_int(actions, i+2)));
    PC("} _data, *data = &_data;\n  ");
    PC(get_string((string)list_get_int(actions, i+3)));
    PC("\n}\n");
    PC("#undef REGISTER_HISTORY\n\n");
  }

  return 0;
}

#define P_(x)							\
  do {								\
    char *p = x;						\
    while (*p) {						\
      if (*p == '\n')						\
        if (fputc('\\', h) == -1) {perror(H);return -1;}	\
      if (fputc(*p, h) == -1) {perror(H);return -1;}		\
      p++;							\
    }								\
  } while (0)

/* 0 ok, -1 error */
int put_load_save(void)
{
  int i;

  /* history load */
  PH("#define HISTORY_LOAD \\\n");

  for (i=0; i<list_size(actions); i+=9) {
    P_("      case ACTION_");
    P_(get_string((string)list_get_int(actions, i)));
    P_(": {\n");
    P_("        struct { ");
    P_(get_string((string)list_get_int(actions, i+2)));
    P_(" } *data;\n");
    P_("        if ((ret = increase_history_data(h, sizeof(*data),\n");
    P_("                                         &h->data[i].size)) < 0)\n");
    P_("          INPUT_ERROR_1(ret);\n");
    P_("        if (major) h->data[i].size = -h->data[i].size;\n");
    P_("        h->data[i].data = ret;\n");
    P_("        data = h->history_data + ret;\n");
    P_(get_string((string)list_get_int(actions, i+7)));
    P_("        break;\n");
    P_("      }\n");
  }

  PH("  /**/\n\n");

  /* history save */
  PH("#define HISTORY_SAVE \\\n");

  for (i=0; i<list_size(actions); i+=9) {
    P_("      case ACTION_");
    P_(get_string((string)list_get_int(actions, i)));
    P_(": {\n");
    P_("        struct { ");
    P_(get_string((string)list_get_int(actions, i+2)));
    P_(" } *data = d;\n");
    P_(get_string((string)list_get_int(actions, i+6)));
    P_("        break;\n");
    P_("      }\n");
  }

  PH("  /**/\n\n");

  return 0;
}

/* 0 ok, -1 error */
int create_actions(void)
{
  int i;

  c = fopen(C, "w"); if (!c) { perror(C); return -1; }
  h = fopen(H, "w"); if (!h) { perror(H); return -1; }

  /* header */
  PC("/* This file was generated by 'l'. ");
  PC("You probably should not modify it. */\n\n");

  PH("/* This file was generated by 'l'. ");
  PH("You probably should not modify it. */\n\n");
  PH("#ifndef _ACTIONS_H_\n#define _ACTIONS_H_\n\n");

  /* enum */
  if (put_enum() == -1) return -1;

  /* includes */
  PC("#include \"actions.h\"\n");
  PC("#include <string.h>\n");
  PH("#include \"typedefs.h\"\n");

  for (i=0; i<list_size(include); i++) {
    PC("#include");
    PC(get_string((string)list_get_int(include, i)));
    PC("\n");

    PH("#include");
    PH(get_string((string)list_get_int(include, i)));
    PH("\n");
  }

  for (i=0; i<list_size(cinclude); i++) {
    PC("#include");
    PC(get_string((string)list_get_int(cinclude, i)));
    PC("\n");
  }

  for (i=0; i<list_size(hinclude); i++) {
    PH("#include");
    PH(get_string((string)list_get_int(hinclude, i)));
    PH("\n");
  }

  PC("\n");
  PH("\n");

  /* functions prototypes */
  if (put_proto() == -1) return -1;

  /* undo, redo and erase stuff */
  if (put_undo() == -1) return -1;
  if (put_redo() == -1) return -1;
  if (put_erase() == -1) return -1;

  /* functions definitions */
  if (put_function() == -1) return -1;

  /* history */
  if (put_history() == -1) return -1;

  /* load/save stuff */
  if (put_load_save() == -1) return -1;

  /* trailer */
  PH("#endif /* _ACTIONS_H_ */\n");

  if (fclose(c) == -1) { perror(C); return -1; }
  if (fclose(h) == -1) { perror(H); return -1; }

  return 0;
}
