/*
 * l: a language for cgscore to describe actions
 * public domain, like cgscore
 */

%{

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#include "keywords.h"
#include "files.h"
#include "string.h"
#include "list.h"
#include "actions.h"

#define LEXER_NORMAL	0
#define LEXER_F_ARGS	1
#define LEXER_INCLUDE	2
#define LEXER_CODE	3
#define LEXER_DATA	4

extern list include;
extern list cinclude;
extern list hinclude;
extern list actions;
extern list registered_actions;

static int lexer_mode = LEXER_NORMAL;
static int line = 0;

static char *buffer;
static int buffer_size;
static int buffer_maxsize;
static int put(int);
void end_parse(void);
static int yylex(void);
void yyerror(char *);

static int data;
static int _do;
static int undo;
static int redo;
static int save;
static int load;
static int erase;

%}

%start file

%token INCLUDE CINCLUDE HINCLUDE ACTION DATA DO UNDO REDO SAVE LOAD F_ARGS
%token IDENTIFIER CODE ERASE

%%

file:
	  includes actions
	;

includes:
	  /* nothing */
	| includes '#' include
	;

include:
	  INCLUDE include_string
	  {
	    if (list_add_int(include, $2) == -1) YYERROR;
	  }
	| CINCLUDE include_string
	  {
	    if (list_add_int(cinclude, $2) == -1) YYERROR;
	  }
	| HINCLUDE include_string
	  {
	    if (list_add_int(hinclude, $2) == -1) YYERROR;
	  }
	;

actions:
	  /* nothing */
	| actions action
	;

action:
	  {
	    data = -1;
	    _do = -1;
	    undo = -1;
	    redo = -1;
	    save = -1;
	    load = -1;
	    erase = -1;
	  }
	  action_declaration '{' action_body '}'
	  {
	    if (data == -1) {
	      yyerror("bad declaration: data missing");
	      YYERROR;
	    }
	    if (_do == -1) {
	      yyerror("bad declaration: do missing");
	      YYERROR;
	    }
	    if (undo == -1) {
	      yyerror("bad declaration: undo missing");
	      YYERROR;
	    }
	    if (redo == -1) {
	      yyerror("bad declaration: redo missing");
	      YYERROR;
	    }
	    if (save == -1) {
	      yyerror("bad declaration: save missing");
	      YYERROR;
	    }
	    if (load == -1) {
	      yyerror("bad declaration: load missing");
	      YYERROR;
	    }
	    if (erase == -1) {
	      yyerror("bad declaration: erase missing");
	      YYERROR;
	    }
	    if (list_add_int(actions, data) == -1 ||
	        list_add_int(actions, _do) == -1 ||
	        list_add_int(actions, undo) == -1 ||
	        list_add_int(actions, redo) == -1 ||
	        list_add_int(actions, save) == -1 ||
	        list_add_int(actions, load) == -1 ||
	        list_add_int(actions, erase) == -1) YYERROR;
	    if (register_action(list_size(actions)-9) == -1) YYERROR;
	  }
	;

action_declaration:
	  ACTION identifier
	  {
	    if (list_add_int(actions, $2) == -1 ||
	        list_add_int(actions, -1) == -1) YYERROR;
	  }
	| ACTION identifier '(' action_declaration_arguments ')'
	  {
	    if (list_add_int(actions, $2) == -1 ||
	        list_add_int(actions, $4) == -1) YYERROR;
	  }
	;

action_body:
	  action_items
	;

action_items:
	  /* nothing */
	| action_items action_item
	;

action_item:
	  DATA '{' data_arguments '}'
	  {
	    if (data != -1) {
	      yyerror("data already defined");
	      YYERROR;
	    }
	    data = $3;
	  }
	| DO '{' ccode '}'
	  {
	    if (_do != -1) {
	      yyerror("do already defined");
	      YYERROR;
	    }
	    _do = $3;
	  }
	| UNDO '{' ccode '}'
	  {
	    if (undo != -1) {
	      yyerror("undo already defined");
	      YYERROR;
	    }
	    undo = $3;
	  }
	| REDO '{' ccode '}'
	  {
	    if (redo != -1) {
	      yyerror("redo already defined");
	      YYERROR;
	    }
	    redo = $3;
	  }
	| SAVE '{' ccode '}'
	  {
	    if (save != -1) {
	      yyerror("save already defined");
	      YYERROR;
	    }
	    save = $3;
	  }
	| LOAD '{' ccode '}'
	  {
	    if (load != -1) {
	      yyerror("load already defined");
	      YYERROR;
	    }
	    load = $3;
	  }
	| ERASE '{' ccode '}'
	  {
	    if (erase != -1) {
	      yyerror("erase already defined");
	      YYERROR;
	    }
	    erase = $3;
	  }
	;

action_declaration_arguments:
	  { lexer_mode = LEXER_F_ARGS; }
	  F_ARGS
	  {
	    string s = new_string(buffer);
	    if (s == -1) YYERROR;
	    lexer_mode = LEXER_NORMAL;
	    $$ = s;
	  }
	;

include_string:
	  { lexer_mode = LEXER_INCLUDE; }
	  INCLUDE
	  {
	    string s = new_string(buffer);
	    if (s == -1) YYERROR;
	    lexer_mode = LEXER_NORMAL;
	    $$ = s;
	  }
	;

identifier:
	  IDENTIFIER
	  {
	    string s = new_string(buffer);
	    if (s == -1) YYERROR;
	    $$ = s;
	  }
	;

ccode:
	  { lexer_mode = LEXER_CODE; }
	  CODE
	  {
	    string s = new_string(buffer);
	    if (s == -1) YYERROR;
	    lexer_mode = LEXER_NORMAL;
	    $$ = s;
	  }
	;

data_arguments:
	  { lexer_mode = LEXER_DATA; }
	  CODE
	  {
	    string s = new_string(buffer);
	    if (s == -1) YYERROR;
	    lexer_mode = LEXER_NORMAL;
	    $$ = s;
	  }
	;

%%

/* return 0 if ok, -1 if error */
int put(int c)
{
  if (buffer_size == buffer_maxsize) {
    char *nb;
    if (!buffer_size)
      nb = malloc(256);
    else
      nb = realloc(buffer, buffer_maxsize + 256);
    if (!nb) {
      fprintf(stderr, "out of memory\n");
      return -1;
    }
    buffer = nb;
    buffer_maxsize += 256;
  }
  buffer[buffer_size] = c;
  buffer_size++;
  return 0;
}

int get_byte(void)
{
  int c = get();
  if (c == EOF) return -1;
  if (c == '\n') line++;
  return c;
}

void unget_byte(int c)
{
  if (c == '\n') line--;
  unget(c);
}

/* -1 error, 0 no comment, 1 comment removed */
int check_comments(int c, int put_it)
{
  if (c != '/') return 0;
  c = get_byte();

  if (c == '*') {
    if (put_it) if (put('/') == -1 || put('*') == -1) return -1;
    while (1) {
      do {
        c = get_byte();
        if (c == -1) { yyerror("bad C comment"); return -1; }
        if (put_it) if (put(c) == -1) return -1;
      } while (c != '*');
      c = get_byte();
      if (c == '/') {
        if (put_it) if (put(c) == -1) return -1;
        return 1;
      }
      unget_byte(c);
    }
  }

  if (c == '/') {
    if (put_it) if (put('/') == -1 || put('/') == -1) return -1;
    while (1) {
      c = get_byte();
      if (put_it && c != -1) if (put(c) == -1) return -1;
      if (c == '\n') return 1;
      if (c == -1) { yyerror("bad C comment"); return -1; }
    }
  }

  unget_byte(c);
  return 0;
}

/* return 0 if ok, -1 if error */
int get_full_string(int sep)
{
  int c = get_byte();
  int last_escape = 0;

  if (c != sep) return -1;

  do {
    if (put(c) == -1) return -1;
    if (last_escape) last_escape = 0; else last_escape = (c == '\\');
    c = get_byte();
    if (c == -1) {
      yyerror("unterminated string sequence");
      return -1;
    }
  } while (c != sep || last_escape);

  if (put(sep) == -1) return -1;

  return 0;
}

static int get_code(int put_it)
{
  int depth = 0;
  if (put_it) if (put('{') == -1) return -1;
  while (1) {
    int c = get_byte();
    int comment;

    if (c == -1) {
      yyerror("bad C code");
      return -1;
    }

    comment = check_comments(c, 1);
    if (comment == -1) return -1;
    if (comment == 1) continue;

    if (c == '}' && depth == 0) {
      unget_byte(c);
      if (put_it) if (put('}') == -1) return -1;
      if (put(0) == -1) return -1;
      return CODE;
    }

    if (put(c) == -1) return -1;

    if (c == '{') depth++;
    if (c == '}') depth--;

    if (c == '"' || c == '\'') {
      buffer_size--;
      unget_byte(c);
      if (get_full_string(c) == -1)
        return -1;
    }
  }
}

int get_include(void)
{
  while (1) {
    int c = get_byte();
    if (c == -1 || c == '\n') {
      if (put(0) == -1) return -1;
      break;
    }
    if (put(c) == -1) return -1;
  }

  return INCLUDE;
}

int get_f_args(void)
{
  int c = get_byte();
  int depth = 0;

  while (c != ')' || depth) {
    int comment = check_comments(c, 0);
    if (comment == -1) return -1;

    if (!comment) if (put(c) == -1) return -1;
    if (c == '(') depth++;
    if (c == ')') depth--;
    if (c == '"' || c == '\'') {
      buffer_size--;
      unget_byte(c);
      if (get_full_string(c) == -1)
        return -1;
    }
    c = get_byte();
    if (c == -1) {
      yyerror("bad function arguments");
      return -1;
    }
  }

  if (put(0) == -1) return -1;

  unget_byte(')');

  return F_ARGS;
}

int yylex(void)
{
  int c;
  int comment;
  keyword k;

lex_more:
  buffer_size = 0;

  if (lexer_mode == LEXER_CODE)
    return get_code(1);

  if (lexer_mode == LEXER_DATA)
    return get_code(0);

  if (lexer_mode == LEXER_F_ARGS)
    return get_f_args();

  if (lexer_mode == LEXER_INCLUDE)
    return get_include();

  /* let's remove spaces */
  while (1) {
    c = get_byte();
    if (c == -1) return -1;
    if (!isspace(c))
      break;
  }

  /* keyword or identifier */
  if ((c >= 'a' && c <= 'z') ||
      (c >= 'A' && c <= 'Z') ||
      c == '_') {
    while (1) {
      if (put(c) == -1) return -1;
      c = get_byte();
      if ((c >= 'a' && c <= 'z') ||
          (c >= 'A' && c <= 'Z') ||
          (c >= '0' && c <= '9') ||
          c == '_')
        continue;
      unget_byte(c);
      if (put(0) == -1) return -1;
      switch ((k=lookup(buffer, buffer_size-1))) {
        case NULL_KEYWORD:
          return IDENTIFIER;
        case KEYWORD_INCLUDE:
          return INCLUDE;
        case KEYWORD_CINCLUDE:
          return CINCLUDE;
        case KEYWORD_HINCLUDE:
          return HINCLUDE;
        case KEYWORD_ACTION:
          return ACTION;
        case KEYWORD_DATA:
          return DATA;
        case KEYWORD_DO:
          return DO;
        case KEYWORD_UNDO:
          return UNDO;
        case KEYWORD_REDO:
          return REDO;
        case KEYWORD_SAVE:
          return SAVE;
        case KEYWORD_LOAD:
          return LOAD;
        case KEYWORD_ERASE:
          return ERASE;
      }
    }
  }

  /* comments */
  comment = check_comments(c, 0);
  if (comment == -1) return -1;
  if (comment == 1) goto lex_more;

  return c;
}

/* 0 ok, -1 error */
int parse_registered_actions(void)
{
  string s;
  int c;
  if (init_file("actions.list") == -1) return -1;
  /* get all registered actions */
  while (1) {
    /* remove spaces (return 0 if end of file (or error) reached) */
    do c = get(); while (isspace(c) && c != -1); if (c == -1) return 0;
    /* comment? */
    if (c == '#') {
      do c = get(); while (c != -1 && c != '\n'); if (c == -1) return -1;
      continue;
    }
    if (!isalpha(c) && c != '_') return -1;
    /* action name, get it full */
    buffer_size = 0;
    do {
      if (put(c) == -1) return -1;
      c = get();
    } while (isalnum(c) || c == '_');
    unget(c);
    if (c == -1) return -1;
    if (put(0) == -1) return -1;
    /* add it to the know actions */
    s = new_string(buffer);
    if (s == -1) return -1;
    if (list_add_int(registered_actions, (int)s) == -1 ||
        list_add_int(registered_actions, -1) == -1)
      return -1;
  }
  close_file();
  return 0;
}

void yyerror(char *s)
{
  fprintf(stderr, "%s:%d:%s\n", get_current_file(), line+1, s);
}

void end_parse(void)
{
  if (!buffer_maxsize) return;
  free(buffer);
  buffer_size = buffer_maxsize = 0;
  buffer = 0;
  lexer_mode = LEXER_NORMAL;
}

void reset_parse(void)
{
  lexer_mode = LEXER_NORMAL;
  line = 0;
}
