/*
 * m: a language for cgscore to describe states
 * public domain, like cgscore
 */

%{

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>

#include "list.h"
#include "keywords.h"
#include "string.h"
#include "files.h"
#include "c_fun_args.h"

#define LEXER_NORMAL	0
#define LEXER_F_ARGS	1
#define LEXER_INCLUDE	2
#define LEXER_CODE	3
#define LEXER_DATA	4

extern list include;
extern list cinclude;
extern list hinclude;
extern list states;

static int lexer_mode = LEXER_NORMAL;
static int line = 0;

static char *buffer;
static int buffer_size;
static int buffer_maxsize;
static int put(int);
void end_parse(void);
static int yylex(void);
void yyerror(char *);

static list current_state;
static string data;
static string enter;
static string leave;
static list actions;

%}

%start file

%token INCLUDE CINCLUDE HINCLUDE STATE DATA ENTER LEAVE ACTION F_ARGS
%token IDENTIFIER CODE

%%

file:
	  includes states
	;

includes:
	  /* nothing */
	| includes '#' include
	;

include:
	  INCLUDE include_string
	  {
	    if (list_add_int(include, $2) == -1) YYERROR;
	  }
	| CINCLUDE include_string
	  {
	    if (list_add_int(cinclude, $2) == -1) YYERROR;
	  }
	| HINCLUDE include_string
	  {
	    if (list_add_int(hinclude, $2) == -1) YYERROR;
	  }
	;

states:
	  /* nothing */
	| states state
	;

state:
	  {
	    current_state = new_list(); if (current_state == -1) YYERROR;
	    data = -1;
	    enter = -1;
	    leave = -1;
	    actions = new_list(); if (actions == -1) YYERROR;
	  }
	  state_declaration '{' state_body '}'
	  {
	    if (data == -1) {
	      yyerror("bad declaration: data missing");
	      YYERROR;
	    }
	    if (enter == -1) {
	      yyerror("bad declaration: enter missing");
	      YYERROR;
	    }
	    if (leave == -1) {
	      yyerror("bad declaration: leave missing");
	      YYERROR;
	    }
	    if (list_add_int(current_state, data) == -1 ||
	        list_add_int(current_state, enter) == -1 ||
	        list_add_int(current_state, leave) == -1 ||
	        list_add_int(current_state, actions) == -1 ||
	        list_add_int(states, current_state) == -1) YYERROR;
	  }
	;

state_declaration:
	  STATE identifier
	  {
	    string s = new_string(get_current_file()); if (s == -1) YYERROR;
	    char *c = get_string(s);
	    /* replace the .m by .c */
	    c[strlen(c)-1] = 'c';
	    if (list_add_int(current_state, $2) == -1 ||
	        list_add_int(current_state, s) == -1 ||
	        list_add_int(current_state, -1) == -1) YYERROR;
	  }
	| STATE identifier '(' state_declaration_arguments ')'
	  {
	    string s = new_string(get_current_file()); if (s == -1) YYERROR;
	    list args = parse_arguments($4);
	    char *c = get_string(s);
	    /* replace the .m by .c */
	    c[strlen(c)-1] = 'c';
	    if (args == -1) {
	      yyerror("error with state's arguments");
	      YYERROR;
	    }
	    if (list_add_int(current_state, $2) == -1 ||
	        list_add_int(current_state, s) == -1 ||
	        list_add_int(current_state, args) == -1) YYERROR;
	  }
	;

state_body:
	  state_items
	;

state_items:
	  /* nothing */
	| state_items state_item
	;

state_item:
	  DATA '{' data_arguments '}'
	  {
	    if (data != -1) {
	      yyerror("data already defined");
	      YYERROR;
	    }
	    data = $3;
	  }
	| ENTER '{' ccode '}'
	  {
	    if (enter != -1) {
	      yyerror("enter already defined");
	      YYERROR;
	    }
	    enter = $3;
	  }
	| LEAVE '{' ccode '}'
	  {
	    if (leave != -1) {
	      yyerror("leave already defined");
	      YYERROR;
	    }
	    leave = $3;
	  }
	| ACTION identifier '{' ccode '}'
	  {
	    list code = parse_code($4);
	    if (code == -1) {
	      yyerror("error with action's code");
	      YYERROR;
	    }
	    if (list_add_int(actions, $2) == -1 ||
	        list_add_int(actions, -1) == -1 ||
	        list_add_int(actions, code) == -1) YYERROR;
	  }
	| ACTION identifier '(' action_declaration_arguments ')' '{' ccode '}'
	  {
	    list args = parse_arguments($4);
	    list code;
	    if (args == -1) {
	      yyerror("error with action's arguments");
	      YYERROR;
	    }
	    code = parse_code($7);
	    if (code == -1) {
	      yyerror("error with action's code");
	      YYERROR;
	    }
	    /* let's "forget" the list if nothing in it */
	    if (list_size(args) <= 1) args = -1;
	    if (list_add_int(actions, $2) == -1 ||
	        list_add_int(actions, args) == -1 ||
	        list_add_int(actions, code) == -1) YYERROR;
	  }
	;

action_declaration_arguments:
	  function_arguments
	  { $$ = $1; }
	;

state_declaration_arguments:
	  function_arguments
	  { $$ = $1; }
	;

function_arguments:
	  { lexer_mode = LEXER_F_ARGS; }
	  F_ARGS
	  {
	    string s = new_string(buffer);
	    if (s == -1) YYERROR;
	    lexer_mode = LEXER_NORMAL;
	    $$ = s;
	  }
	;

include_string:
	  { lexer_mode = LEXER_INCLUDE; }
	  INCLUDE
	  {
	    string s = new_string(buffer);
	    if (s == -1) YYERROR;
	    lexer_mode = LEXER_NORMAL;
	    $$ = s;
	  }
	;

identifier:
	  IDENTIFIER
	  {
	    string s = new_string(buffer);
	    if (s == -1) YYERROR;
	    $$ = s;
	  }
	;

ccode:
	  { lexer_mode = LEXER_CODE; }
	  CODE
	  {
	    string s = new_string(buffer);
	    if (s == -1) YYERROR;
	    lexer_mode = LEXER_NORMAL;
	    $$ = s;
	  }
	;

data_arguments:
	  { lexer_mode = LEXER_DATA; }
	  CODE
	  {
	    string s = new_string(buffer);
	    if (s == -1) YYERROR;
	    lexer_mode = LEXER_NORMAL;
	    $$ = s;
	  }
	;

%%

/* return 0 if ok, -1 if error */
int put(int c)
{
  if (buffer_size == buffer_maxsize) {
    char *nb;
    if (!buffer_size)
      nb = malloc(256);
    else
      nb = realloc(buffer, buffer_maxsize + 256);
    if (!nb) {
      fprintf(stderr, "out of memory\n");
      return -1;
    }
    buffer = nb;
    buffer_maxsize += 256;
  }
  buffer[buffer_size] = c;
  buffer_size++;
  return 0;
}

int get_byte(void)
{
  int c = get();
  if (c == EOF) return -1;
  if (c == '\n') line++;
  return c;
}

void unget_byte(int c)
{
  if (c == '\n') line--;
  unget(c);
}

/* -1 error, 0 no comment, 1 comment removed */
int check_comments(int c, int put_it)
{
  if (c != '/') return 0;
  c = get_byte();

  if (c == '*') {
    if (put_it) if (put('/') == -1 || put('*') == -1) return -1;
    while (1) {
      do {
        c = get_byte();
        if (c == -1) { yyerror("bad C comment"); return -1; }
        if (put_it) if (put(c) == -1) return -1;
      } while (c != '*');
      c = get_byte();
      if (c == '/') {
        if (put_it) if (put(c) == -1) return -1;
        return 1;
      }
      unget_byte(c);
    }
  }

  if (c == '/') {
    if (put_it) if (put('/') == -1 || put('/') == -1) return -1;
    while (1) {
      c = get_byte();
      if (put_it && c != -1) if (put(c) == -1) return -1;
      if (c == '\n') return 1;
      if (c == -1) { yyerror("bad C comment"); return -1; }
    }
  }

  unget_byte(c);
  return 0;
}

/* return 0 if ok, -1 if error */
int get_full_string(int sep)
{
  int c = get_byte();
  int last_escape = 0;
  int start_line = line;

  if (c != sep) return -1;

  do {
    if (put(c) == -1) return -1;
    if (last_escape) last_escape = 0; else last_escape = (c == '\\');
    c = get_byte();
    if (c == '\n') {
      static int done = 0;
      if (!done) {
        done = 1;
        fprintf(stderr, "%s:%d:warning: end-of-line in string constant\n",
                get_current_file(), line);
      }
    }
    if (c == -1) {
      yyerror("unterminated string sequence");
      fprintf(stderr, "  (possible starting line: %d)\n", start_line+1);
      return -1;
    }
  } while (c != sep || last_escape);

  if (put(sep) == -1) return -1;

  return 0;
}

static int get_code(int put_it)
{
  int depth = 0;
  if (put_it) if (put('{') == -1) return -1;
  while (1) {
    int c = get_byte();
    int comment;

    if (c == -1) {
      yyerror("bad C code");
      return -1;
    }

    comment = check_comments(c, 1);
    if (comment == -1) return -1;
    if (comment == 1) continue;

    if (c == '}' && depth == 0) {
      unget_byte(c);
      if (put_it) if (put('}') == -1) return -1;
      if (put(0) == -1) return -1;
      return CODE;
    }

    if (put(c) == -1) return -1;

    if (c == '{') depth++;
    if (c == '}') depth--;

    if (c == '"' || c == '\'') {
      buffer_size--;
      unget_byte(c);
      if (get_full_string(c) == -1)
        return -1;
    }
  }
}

int get_include(void)
{
  while (1) {
    int c = get_byte();
    if (c == -1 || c == '\n') {
      if (put(0) == -1) return -1;
      break;
    }
    if (put(c) == -1) return -1;
  }

  return INCLUDE;
}

int get_f_args(void)
{
  int c = get_byte();
  int depth = 0;

  while (c != ')' || depth) {
    int comment = check_comments(c, 0);
    if (comment == -1) return -1;

    if (!comment) if (put(c) == -1) return -1;
    if (c == '(') depth++;
    if (c == ')') depth--;
    if (c == '"' || c == '\'') {
      buffer_size--;
      unget_byte(c);
      if (get_full_string(c) == -1)
        return -1;
    }
    c = get_byte();
    if (c == -1) {
      yyerror("bad function arguments");
      return -1;
    }
  }

  if (put(0) == -1) return -1;

  unget_byte(')');

  return F_ARGS;
}

int yylex(void)
{
  int c;
  int comment;
  keyword k;

lex_more:
  buffer_size = 0;

  if (lexer_mode == LEXER_CODE)
    return get_code(1);

  if (lexer_mode == LEXER_DATA)
    return get_code(0);

  if (lexer_mode == LEXER_F_ARGS)
    return get_f_args();

  if (lexer_mode == LEXER_INCLUDE)
    return get_include();

  /* let's remove spaces */
  while (1) {
    c = get_byte();
    if (c == -1) return -1;
    if (!isspace(c))
      break;
  }

  /* keyword or identifier */
  if ((c >= 'a' && c <= 'z') ||
      (c >= 'A' && c <= 'Z') ||
      c == '_') {
    while (1) {
      if (put(c) == -1) return -1;
      c = get_byte();
      if ((c >= 'a' && c <= 'z') ||
          (c >= 'A' && c <= 'Z') ||
          (c >= '0' && c <= '9') ||
          c == '_')
        continue;
      unget_byte(c);
      if (put(0) == -1) return -1;
      switch ((k=lookup(buffer, buffer_size-1))) {
        case NULL_KEYWORD:
          return IDENTIFIER;
        case KEYWORD_INCLUDE:
          return INCLUDE;
        case KEYWORD_CINCLUDE:
          return CINCLUDE;
        case KEYWORD_HINCLUDE:
          return HINCLUDE;
        case KEYWORD_STATE:
          return STATE;
        case KEYWORD_DATA:
          return DATA;
        case KEYWORD_ENTER:
          return ENTER;
        case KEYWORD_LEAVE:
          return LEAVE;
        case KEYWORD_ACTION:
          return ACTION;
      }
    }
  }

  /* comments */
  comment = check_comments(c, 0);
  if (comment == -1) return -1;
  if (comment == 1) goto lex_more;

  return c;
}

void yyerror(char *s)
{
  fprintf(stderr, "%s:%d:%s\n", get_current_file(), line+1, s);
}

void end_parse(void)
{
  if (!buffer_maxsize) return;
  free(buffer);
  buffer_size = buffer_maxsize = 0;
  buffer = 0;
  lexer_mode = LEXER_NORMAL;
}

void reset_parse(void)
{
  lexer_mode = LEXER_NORMAL;
  line = 0;
}
