/*
 * m: a language for cgscore to describe states
 * public domain, like cgscore
 */

#include "list.h"

#include <stdlib.h>
#include <stdio.h>

typedef union {
  void *p;
  int i;
  float f;
} _list_elem;

typedef struct {
  _list_elem *buffer;
  int size;
  int maxsize;
} _list;

static _list *buffer;
static int size;
static int maxsize;

/* -1 : error, list : ok */
list new_list(void)
{
  if (size == maxsize) {
    int new_size = maxsize + 64;
    _list *new;
    if (!maxsize)
      new = malloc(new_size * sizeof(_list));
    else
      new = realloc(buffer, new_size * sizeof(_list));
    if (!new) {
      fprintf(stderr, "out of memory\n");
      return -1;
    }
    maxsize = new_size;
    buffer = new;
  }

  buffer[size].maxsize = buffer[size].size = 0;
  buffer[size].buffer = 0;
  size++;

  return size-1;
}

int list_grow(_list *l)
{
  if (l->size == l->maxsize) {
    int new_size = l->maxsize + 64;
    _list_elem *new;
    if (!l->maxsize)
      new = malloc(new_size * sizeof(_list_elem));
    else
      new = realloc(l->buffer, new_size * sizeof(_list_elem));
    if (!new) {
      fprintf(stderr, "out of memory\n");
      return -1;
    }
    l->maxsize = new_size;
    l->buffer = new;
  }
  return 0;
}

/* -1 error, 0 ok */
int list_add(list ln, void *data)
{
  _list *l = &buffer[ln];

  if (list_grow(l) == -1) return -1;

  l->buffer[l->size].p = data;
  l->size++;

  return 0;
}

/* -1 error, 0 ok */
int list_add_int(list ln, int data)
{
  _list *l = &buffer[ln];

  if (list_grow(l) == -1) return -1;

  l->buffer[l->size].i = data;
  l->size++;

  return 0;
}

/* -1 error, 0 ok */
int list_add_float(list ln, float data)
{
  _list *l = &buffer[ln];

  if (list_grow(l) == -1) return -1;

  l->buffer[l->size].f = data;
  l->size++;

  return 0;
}

void free_list(_list *l)
{
  if (!l->maxsize) return;

  free(l->buffer);
  l->size = l->maxsize = 0;
  l->buffer = 0;
}

void close_list(void)
{
  int i;

  for (i=0; i<size; i++)
    free_list(&buffer[i]);

  if (!maxsize) return;

  free(buffer);
  size = maxsize = 0;
  buffer = 0;
}

void *list_get(list l, int i)
{
  return buffer[l].buffer[i].p;
}

int list_get_int(list l, int i)
{
  return buffer[l].buffer[i].i;
}

void list_set_int(list l, int i, int v)
{
  buffer[l].buffer[i].i = v;
}

float list_get_float(list l, int i)
{
  return buffer[l].buffer[i].f;
}

int list_size(list l)
{
  if (l == -1) return 0;
  return buffer[l].size;
}
