/*
 * m: a language for cgscore to describe states
 * public domain, like cgscore
 */

#include "states.h"

#include <string.h>
#include <stdio.h>
#include <ctype.h>

#include "list.h"
#include "debug.h"
#include "string.h"

/* barbarian external globals (reminds me my daily life of corporate coder) */
extern list include;
extern list cinclude;
extern list hinclude;
extern list states;

extern int only_c;
extern int verbose;
extern int split;

char *hout;

#define Vprintf if (verbose) fprintf

#define STATE_ID(s) (string)list_get_int((list)list_get_int(states, s), 0)
#define STATE_FILENAME(s) \
  (string)list_get_int((list)list_get_int(states, s), 1)
#define STATE_ARGUMENTS(s) \
  (list)list_get_int((list)list_get_int(states, s), 2)
#define STATE_DATA(s) (string)list_get_int((list)list_get_int(states, s), 3)
#define STATE_ENTER(s) (string)list_get_int((list)list_get_int(states, s), 4)
#define STATE_LEAVE(s) (string)list_get_int((list)list_get_int(states, s), 5)
#define STATE_ACTIONS(s) (list)list_get_int((list)list_get_int(states, s), 6)

#define STATE_NB_ACTIONS(s) (list_size(STATE_ACTIONS(s))/3)
#define ACTION_ID(s,i) (string)list_get_int(STATE_ACTIONS(s), (i)*3)
#define ACTION_ARGUMENTS(s,i) (list)list_get_int(STATE_ACTIONS(s), (i)*3+1)
#define ACTION_CODE(s,i) (list)list_get_int(STATE_ACTIONS(s), (i)*3+2)

#define ARGUMENTS_STRING(a) (string)list_get_int(a, 0)
#define NB_ARGUMENTS(a) (a == -1 ? 0 : (list_size(a)-1)/4)
#define ARGUMENT_START(a, i) list_get_int(a, 1+(i)*4)
#define ARGUMENT_END(a, i) list_get_int(a, 2+(i)*4)
#define ARGUMENT_NAME(a, i) list_get_int(a, 3+(i)*4)
#define ARGUMENT_NAME_END(a, i) list_get_int(a, 4+(i)*4)

#define CODE_STRING(c) (string)list_get_int(c, 0)
#define CODE_NB_CALLS(c) ((list_size(c)-1)/2)
#define CODE_CALL_START(c, i)  list_get_int(c, (i)*2+1)
#define CODE_CALL_END(c, i) list_get_int(c, (i)*2+2)

/* 0 ok, -1 error */
static int check_state(int s)
{
  int i;
  int j;
  char *id_i;
  char *id_j;
  int ret = 0;

  Vprintf(stderr, ".. entering %s\n", __FUNCTION__);

  /* all actions in a given state shall be unique */
  SLOW;
  for (i = 0; i < STATE_NB_ACTIONS(s) - 1; i++)
    for (j = i + 1; j < STATE_NB_ACTIONS(s); j++) {
      id_i = get_string(ACTION_ID(s, i));
      id_j = get_string(ACTION_ID(s, j));
      if (!strcmp(id_i, id_j)) {
        fprintf(stderr, "error: state %s defines action %s more than once\n",
          get_string(STATE_ID(s)), id_i);
        ret = -1;
      }
    }

  return ret;
}

/* 0 ok, -1 error */
int check_states(void)
{
  int i;
  int j;
  char *id_i;
  char *id_j;
  int ret = 0;

  Vprintf(stderr, ".. entering %s\n", __FUNCTION__);

  /* all state shall have a unique name */
  SLOW;
  for (i = 0; i < list_size(states) - 1; i++)
    for (j = i + 1; j < list_size(states); j++) {
      id_i = get_string(STATE_ID(i));
      id_j = get_string(STATE_ID(j));
      if (!strcmp(id_i, id_j)) {
        fprintf(stderr, "error: state %s defined more than once\n", id_i);
        ret = -1;
      }
    }

  /* if two states have same name, don't bother testing more */
  if (ret == -1) return -1;

  /* all actions in a given state shall be unique */
  for (i = 0; i < list_size(states); i++)
    if (check_state(i) == -1) ret = -1;

  return ret;
}

#define C "states.c"
static char *H = "states.h";
static char *CC;
#define PC(x) do \
  { if (fprintf(c, "%s", x) == -1) {perror(C);return -1;}} while(0)
#define PH(x) do if (!only_c) \
  { if (fprintf(h, "%s", x) == -1) {perror(H);return -1;}} while(0)
#define PCC(x) do \
  { if (fprintf(cc, "%s", x) == -1) {perror(CC);return -1;}} while(0)

/* ugly globals - I work(ed) for a corporate company you know... */
static FILE *c, *h, *cc;

/* 0: action not already done, 1: already done
 * this should be optimized
 * ideally: every string should be unique all over the program
 *          and we could have some available bits associated to
 *          the string so that O(1) to check wether we already
 *          saw an action or not
 *          or if we don't touch the string stuff, use a hashtable to
 *          store already seen actions
 * laziness you know
 */
int action_already_seen(int cur_state, char *action)
{
  int s;
  int i;

  Vprintf(stderr, ".. entering %s\n", __FUNCTION__);

  /* check that all already done states don't have the action in them
   * if one of them does, it means we already saw the action
   */
  SLOW;
  for (s = 0; s < cur_state; s++) {
    for (i = 0; i < STATE_NB_ACTIONS(s); i++)
      if (!strcmp(get_string(ACTION_ID(s, i)), action)) return 1;
  }

  return 0;
}

/* 0 ok, -1 error */
static int put_state_struct_action(int s)
{
  int i;

  Vprintf(stderr, ".. entering %s\n", __FUNCTION__);

  for (i = 0; i < STATE_NB_ACTIONS(s); i++) {
    /* we put the name of the action only once */
    if (action_already_seen(s, get_string(ACTION_ID(s,i)))) continue;

    PH("  int (*");
    PH(get_string(ACTION_ID(s, i)));
    PH(")(thread, ...);\n");
  }

  return 0;
}

/* 0 ok, -1 error */
static int put_state_actions(int s)
{
  int i;

  Vprintf(stderr, ".. entering %s\n", __FUNCTION__);

  PC("  /* _m_type_");
  PC(get_string(STATE_ID(s)));
  PC(" */\n");
  PC("  {\n");

  for (i = 0; i < STATE_NB_ACTIONS(s); i++) {
    PC("    .");
    PC(get_string(ACTION_ID(s, i)));
    PC(" = ");
    PC(get_string(STATE_ID(s)));
    PC("_");
    PC(get_string(ACTION_ID(s, i)));
    PC(",\n");
  }

  PC("  },\n");

  return 0;
}

/* 0 ok, -1 error */
static int put_struct_actions(void)
{
  int i;
  char buf[32];

  Vprintf(stderr, ".. entering %s\n", __FUNCTION__);

  PH("struct actions {\n");

  for (i = 0; i < list_size(states); i++)
    put_state_struct_action(i);

  PH("};\n\n");

  PH("extern struct actions actions[");
  sprintf(buf, "%d", list_size(states)+1);
  PH(buf);
  PH("];\n\n");

  return 0;
}

/* 0 ok, -1 error */
/* a declaration of variable is:
 * <type> <name> like "int x"
 * it could also be
 * <type> <name> <more type> like "int x[12]" or "float (*f)(int)"
 * so what matters is the position of the name
 * we put everything before, then the name we want (__XX) then
 * everything after the name
 */
int put_argument_type_with_name(list args, int n)
{
  int i;
  char c[32];
  char *s = get_string(ARGUMENTS_STRING(args));
  int start = ARGUMENT_START(args, n);
  int start_name = ARGUMENT_NAME(args, n);
  int end_name = ARGUMENT_NAME_END(args, n);
  int end = ARGUMENT_END(args, n);

  Vprintf(stderr, ".. entering %s\n", __FUNCTION__);

  /* put what's before the name */
  c[1] = 0;
  for (i = start; i < start_name; i++) {
    c[0] = s[i];
    PH(c);
  }

  /* put the name we want */
  sprintf(c, "__%d", n+1);
  PH(c);

  /* put everything after the name */
  c[1] = 0;
  for (i = end_name+1; i <= end; i++) {
    c[0] = s[i];
    PH(c);
  }

  return 0;
}

/* 0 ok, -1 error */
static int put_action_macro_do(string name, list args, int check_args)
{
  int i;
  char c[32];
  char *str = get_string(name);
  int len = strlen(str);

  Vprintf(stderr, ".. entering %s\n", __FUNCTION__);

  /* special case: don't check args, but no arg to check anyway */
  if (!check_args && !NB_ARGUMENTS(args)) {
    PH("#define _DO_");
    c[1] = 0;
    for (i = 0; i < len; i++) {
      char up = str[i];
      if (up >= 'a' && up <= 'z') up += 'A' - 'a';
      c[0] = up;
      PH(c);
    }
    PH(" DO_");
    for (i = 0; i < len; i++) {
      char up = str[i];
      if (up >= 'a' && up <= 'z') up += 'A' - 'a';
      c[0] = up;
      PH(c);
    }
    PH("\n\n");
    return 0;
  }

  PH("#define ");
  if (!check_args) PH("_");
  PH("DO_");
  c[1] = 0;
  for (i = 0; i < len; i++) {
    char up = str[i];
    if (up >= 'a' && up <= 'z') up += 'A' - 'a';
    c[0] = up;
    PH(c);
  }
  PH("(_0");
  for (i = 0; i < NB_ARGUMENTS(args); i++) {
    sprintf(c, ", _%d", i+1);
    PH(c);
  }
  PH(") \\\n");

  PH("  do { \\\n");
  PH("    struct _m_state *_m = thread_top(_0); \\\n");
  PH("    int _ret; \\\n");

  if (check_args && NB_ARGUMENTS(args)) {
    PH("    if (0) { \\\n");
    PH("      /* some non-runnable code to get warnings from gcc \\\n");
    PH("       * (and maybe other compilers) \\\n");
    PH("       */ \\\n");
    for (i = 0; i < NB_ARGUMENTS(args); i++) {
      PH("      ");
      if (put_argument_type_with_name(args, i) == -1) return -1;
      PH( " = ");
      sprintf(c, "_%d", i+1);
      PH(c);
      PH("; \\\n");
    }
    PH("    } \\\n");
  }

  PH("    if (actions[_m->type].");
  PH(get_string(name));
  PH(") { \\\n");

  PH("      _m->_case_fun = NULL; \\\n");

  PH("      _ret = actions[_m->type].");
  PH(get_string(name));
  PH("(_0");
  for (i = 0; i < NB_ARGUMENTS(args); i++) {
    sprintf(c, ", _%d", i+1);
    PH(c);
  }
  PH("); \\\n");

  PH("      if (_ret < 0) printf(\"action_error(_0, _ret);(%d,");
  PH(" %d)\\n\", _0, _ret); \\\n");
  PH("    } \\\n");
  PH("  } while (0)\n");

  PH("\n");

  return 0;
}

/* 0 ok, -1 error */
static int put_action_macro(string name, list args)
{
  int i;
  char c[32];
  char *str = get_string(name);
  int len = strlen(str);

  Vprintf(stderr, ".. entering %s\n", __FUNCTION__);

  PH("#define CAN_");
  c[1] = 0;
  for (i = 0; i < len; i++) {
    char up = str[i];
    if (up >= 'a' && up <= 'z') up += 'A' - 'a';
    c[0] = up;
    PH(c);
  }
  PH("(_0) \\\n");
  PH("  (actions[thread_top_type(_0)].");
  PH(str);
  PH(")\n\n");

  put_action_macro_do(name, args, 1);
  put_action_macro_do(name, args, 0);

  return 0;
}

/* 0 ok, -1 error */
static int put_state_action_macros(int s)
{
  int i;

  Vprintf(stderr, ".. entering %s\n", __FUNCTION__);

  for (i = 0; i < STATE_NB_ACTIONS(s); i++) {
    /* we put the action's macro only once */
    if (action_already_seen(s, get_string(ACTION_ID(s,i)))) continue;
    if (put_action_macro(ACTION_ID(s, i),
                         ACTION_ARGUMENTS(s, i)) == -1) return -1;
  }

  return 0;
}

/* 0 ok, -1 error */
static int put_action_macros(void)
{
  int i;

  Vprintf(stderr, ".. entering %s\n", __FUNCTION__);

  for (i = 0; i < list_size(states); i++)
    put_state_action_macros(i);

  return 0;
}

/* 0 ok, -1 error */
int put_state_data(int s)
{
  int i;
  int j;
  list code;
  char enum_value[64];

  Vprintf(stderr, ".. entering %s\n", __FUNCTION__);

  PH("\n");
  PH("struct ");
  PH(get_string(STATE_ID(s)));
  PH("_data {\n");

  PH("  union {\n");

  for (i = 0; i < STATE_NB_ACTIONS(s); i++) {
    PH("    enum {\n");

    code = ACTION_CODE(s, i);
    for (j = 0; j < CODE_NB_CALLS(code)+1; j++) {
      PH("      ");
      sprintf(enum_value, "_m_%d_%d_%d", s, i, j);
      PH(enum_value);
      /* the if is not necessary, but let's be anal (and thus a bit slower) */
      if (j != CODE_NB_CALLS(code)) PH(",");
      PH("\n");
    }

    PH("    } ");
    PH(get_string(ACTION_ID(s, i)));
    PH(";\n");
  }

  PH("  } _case;\n");

  PH(get_string(STATE_DATA(s)));

  PH("\n};\n");

  return 0;
}

/* 0 ok, -1 error */
int put_enum_structs(void)
{
  int i;

  Vprintf(stderr, ".. entering %s\n", __FUNCTION__);

  /* the enum */
  PH("enum _m_state_type {\n");
  PH("  _m_type_0_state,\n");

  for (i = 0; i < list_size(states); i++) {
    PH("  _m_type_");
    PH(get_string(STATE_ID(i)));
    /* the if is not necessary, but let's be anal (and thus a bit slower) */
    if (i != list_size(states)-1) PH(",");
    PH("\n");
  }

  PH("};\n");

  /* the data of each state */
  for (i = 0; i < list_size(states); i++)
    if (put_state_data(i) == -1) return -1;

  /* the struct state */
  PH("\n");
  PH("struct _m_state {\n");
  PH("  thread thread;\n");
  PH("  enum _m_state_type type;\n");
  PH("  int (*_case_fun)(thread, ...);\n");
  PH("  struct _m_state *next;\n");
  PH("  struct _m_state *previous;\n");
  PH("  union {\n");

  for (i = 0; i < list_size(states); i++) {
    PH("    struct ");
    PH(get_string(STATE_ID(i)));
    PH("_data ");
    PH(get_string(STATE_ID(i)));
    PH(";\n");
  }

  PH("  } u;\n");
  PH("};\n");

  return 0;
}

/* 0 ok, -1 error */
int put_enter(int s)
{
  Vprintf(stderr, ".. entering %s\n", __FUNCTION__);

  PCC("\n");
  PCC("int ");
  PCC(get_string(STATE_ID(s)));
  PCC("_enter(thread _owner");

  PH("\n");
  PH("int ");
  PH(get_string(STATE_ID(s)));
  PH("_enter(thread");

  if (STATE_ARGUMENTS(s) != -1) {
    PCC(", ");
    PCC(get_string(ARGUMENTS_STRING(STATE_ARGUMENTS(s))));
    PH(", ");
    PH(get_string(ARGUMENTS_STRING(STATE_ARGUMENTS(s))));
  }

  PH(");\n");

  PCC(")\n");
  PCC("{\n");
  PCC("  struct _m_state *this = new_state(_m_type_");
  PCC(get_string(STATE_ID(s)));
  PCC(");\n");
  PCC("  struct ");
  PCC(get_string(STATE_ID(s)));
  PCC("_data *data;\n");
  PCC("\n");
  PCC("  if (!this) return -OOM;\n");
  PCC("  thread_push(_owner, this);\n");
  PCC("\n");
  PCC("  data = &this->u.");
  PCC(get_string(STATE_ID(s)));
  PCC(";\n");
  PCC("\n");
  PCC("  ");
  PCC(get_string(STATE_ENTER(s)));
  PCC("\n");
  PCC("\n");
  PCC("  return 0;\n");
  PCC("}\n");

  return 0;
}

/* 0 ok, -1 error */
int put_enters(void)
{
  int i;

  Vprintf(stderr, ".. entering %s\n", __FUNCTION__);

  for (i = 0; i < list_size(states); i++) {
    if (split) {
      CC = get_string(STATE_FILENAME(i));
      cc = fopen(CC, "a"); if (!cc) { perror(CC); return -1; }
    }
    if (put_enter(i) == -1) return -1;
    if (split) {
      if (fclose(cc) == -1) { perror(CC); return -1; }
    }
  }

  return 0;
}

/* 0 ok, -1 error */
int put_leave(int s)
{
  Vprintf(stderr, ".. entering %s\n", __FUNCTION__);

  PCC("\n");
  PCC("int ");
  PCC(get_string(STATE_ID(s)));
  PCC("_leave(thread _owner)\n");
  PCC("{\n");
  PCC("  struct _m_state *this = thread_pop(_owner);\n");
  PCC("  struct ");
  PCC(get_string(STATE_ID(s)));
  PCC("_data *data;\n");
  PCC("  if (!this) return -STATE_POP;\n");
  PCC("  data = &this->u.");
  PCC(get_string(STATE_ID(s)));
  PCC(";\n");
  PCC("\n");
  PCC("  ");
  PCC(get_string(STATE_LEAVE(s)));
  PCC("\n");
  PCC("\n");
  PCC("  free(this);\n");
  PCC("\n");
  PCC("  return 0;\n");
  PCC("}\n");

  return 0;
}

/* 0 ok, -1 error */
int put_leaves(void)
{
  int i;

  Vprintf(stderr, ".. entering %s\n", __FUNCTION__);

  for (i = 0; i < list_size(states); i++) {
    if (split) {
      CC = get_string(STATE_FILENAME(i));
      cc = fopen(CC, "a"); if (!cc) { perror(CC); return -1; }
    }
    if (put_leave(i) == -1) return -1;
    if (split) {
      if (fclose(cc) == -1) { perror(CC); return -1; }
    }
  }

  return 0;
}

/* 0 ok, -1 error */
int put_define_call(int state, int action, int code)
{
  int i;
  int called_state;
  list arguments;
  list l = ACTION_CODE(state, action);
  char *s = get_string(CODE_STRING(l));
  int start_state = CODE_CALL_START(l, code);
  int end_state;
  char buf[64];

  Vprintf(stderr, ".. entering %s\n", __FUNCTION__);
  Vprintf(stderr, ".. code=\n%s\n", s);

  /* we have to find what state is called */
  /* dig for '(' */
  while (s[start_state] && s[start_state] != '(') start_state++;
  if (s[start_state] == 0) goto dc_err;
  start_state++;
  /* remove white space */
  while (s[start_state] && isspace(s[start_state])) start_state++;
  if (s[start_state] == 0) goto dc_err;

  /* find end of state */
  end_state = start_state;
  while (s[end_state] && (s[end_state] == '_' || isalnum(s[end_state])))
    end_state++;
  if (s[end_state] == 0) goto dc_err;
  end_state--;
  if (end_state < start_state) goto dc_err;

  /* find what state is called */
  SLOW;
  for (called_state = 0; called_state < list_size(states); called_state++)
    if (!strncmp(s+start_state, get_string(STATE_ID(called_state)),
                 end_state-start_state+1))
      goto dc_found;
  goto dc_err;

dc_found:
  arguments = STATE_ARGUMENTS(called_state);

  PCC("#undef CALL\n");
  PCC("#define CALL(_0");

  for (i = 0; i < NB_ARGUMENTS(arguments); i++) {
    sprintf(buf, ", _%d", i+1);
    PCC(buf);
  }

  PCC(") \\\n");
  PCC("  do { \\\n");
  PCC("    int ret; \\\n");
  PCC("    ret = _0 ## _enter(_owner");

  for (i = 0; i < NB_ARGUMENTS(arguments); i++) {
    sprintf(buf, ", _%d", i+1);
    PCC(buf);
  }

  PCC("); \\\n");
  PCC("    if (ret < 0) return ret; \\\n");
  PCC("    data->_case.");
  PCC(get_string(ACTION_ID(state, action)));
  PCC(" = _m_");

  sprintf(buf, "%d_%d_%d", state, action, code+1);
  PCC(buf);

  PCC("; \\\n");
  PCC("    this->_case_fun = ");
  PCC(get_string(STATE_ID(state)));
  PCC("_");
  PCC(get_string(ACTION_ID(state, action)));
  PCC("; \\\n");
  PCC("    return 0; \\\n");
  PCC("  } while (0)\n");

  Vprintf(stderr, ".. leaving %s\n", __FUNCTION__);

  return 0;

dc_err:
  fprintf(stderr, "error with calling sequence in the following code\n");
  fprintf(stderr, "(did you put the word 'CALL' in a comment?):\n");
  fprintf(stderr, "%s\n", s);
  return -1;
}

/* 0 ok, -1 error */
int put_substring(char *s, int start, int end)
{
  int i;
  char buf[2];

  Vprintf(stderr, ".. entering %s\n", __FUNCTION__);

  buf[1] = 0;

  /* ugly, I agree */
  for (i = start; i <= end; i++) {
    buf[0] = s[i];

    Vprintf(stderr, ".. turn %d (%c)\n", i, s[i]);

    PCC(buf);
  }

  return 0;
}

/* 0 ok, -1 error */
int put_va_arg(int state, int action)
{
  list args = ACTION_ARGUMENTS(state, action);
  char *s;
  int i;

  Vprintf(stderr, ".. entering %s\n", __FUNCTION__);

  if (args == -1) return 0;

  s = get_string(ARGUMENTS_STRING(args));

  for (i = 0; i < NB_ARGUMENTS(args); i++) {
    PCC("      ");
    if (put_substring(s, ARGUMENT_START(args, i),
                      ARGUMENT_END(args, i)) == -1) return -1;
    PCC(";\n");
  }

  PCC("      va_list _map;\n");
  PCC("      va_start(_map, _owner);\n");

  for (i = 0; i < NB_ARGUMENTS(args); i++) {
    PCC("      ");
    if (put_substring(s, ARGUMENT_NAME(args, i),
                      ARGUMENT_NAME_END(args, i)) == -1) return -1;
    PCC(" = va_arg(_map, ");
    /* we get the type without the name */
    if (put_substring(s, ARGUMENT_START(args, i),
                      ARGUMENT_NAME(args, i) - 1) == -1) return -1;
    if (put_substring(s, ARGUMENT_NAME_END(args, i) + 1,
                      ARGUMENT_END(args, i)) == -1) return -1;
    PCC(");\n");
  }

  PCC("      va_end(_map);\n");

  Vprintf(stderr, ".. leaving %s\n", __FUNCTION__);

  return 0;
}

/* 0 ok, -1 error */
int put_code(int state, int action, int code)
{
  list l = ACTION_CODE(state, action);
  char *s = get_string(CODE_STRING(l));
  int start;
  int end;

  Vprintf(stderr, ".. entering %s\n", __FUNCTION__);
  Vprintf(stderr, ".. code (len=%d) =\n%s\n", strlen(s), s);

  /* special case: no call */
  if (CODE_NB_CALLS(l) == 0) {
    PCC(s);
    return 0;
  }

  start = code ? CODE_CALL_END(l, code-1) + 1 : 0;
  end = code == CODE_NB_CALLS(l) ?
          strlen(s) - 1 : CODE_CALL_END(l, code);

  Vprintf(stderr, ".. start=%d end=%d\n", start, end);

  if (put_substring(s, start, end) == -1) return -1;

  return 0;
}

/* 0 ok, -1 error */
int put_action(int state, int action)
{
  int i;
  list code = ACTION_CODE(state, action);
  char buf[64];

  Vprintf(stderr, ".. entering %s\n", __FUNCTION__);

  if (split) {
    PC("int ");
    PC(get_string(STATE_ID(state)));
    PC("_");
    PC(get_string(ACTION_ID(state, action)));
    PC("(thread _owner, ...);\n");
  }

  PCC("\n#undef RETURN\n");
  PCC("#define RETURN \\\n");
  PCC("  do { \\\n");
  PCC("    struct _m_state *_m; \\\n");
  PCC("    int ret; \\\n");
  PCC("    if ((ret = ");
  PCC(get_string(STATE_ID(state)));
  PCC("_leave(_owner)) < 0) return ret; \\\n");
  PCC("    _m = thread_top(_owner); \\\n");
  PCC("    if (!_m) { \\\n");
  PCC("      thread_delete(_owner); \\\n");
  PCC("      return 0; \\\n");
  PCC("    } \\\n");
  PCC("    return _m->_case_fun(_owner); \\\n");
  PCC("  } while (0)\n");

  PCC("#undef SUCCESS\n");
  PCC("#define SUCCESS \\\n");
  PCC("  do { \\\n");
  PCC("    data->_case.");
  PCC(get_string(ACTION_ID(state, action)));
  PCC(" = _m_");

  sprintf(buf, "%d_%d", state, action);
  PCC(buf);

  PCC("_0; \\\n");
  PCC("    return 0; \\\n");
  PCC("  } while (0)\n");

  PCC("#undef ERROR\n");
  PCC("#define ERROR(code) \\\n");
  PCC("  do { \\\n");
  PCC("    data->_case.");
  PCC(get_string(ACTION_ID(state, action)));
  PCC(" = _m_");
  PCC(buf);
  PCC("_0; \\\n");
  PCC("    return code; \\\n");
  PCC("  } while (0)\n");

  PCC("int ");
  PCC(get_string(STATE_ID(state)));
  PCC("_");
  PCC(get_string(ACTION_ID(state, action)));
  PCC("(thread _owner, ...)\n");
  PCC("{\n");
  PCC("  struct _m_state *this = thread_top(_owner);\n");
  PCC("  struct ");
  PCC(get_string(STATE_ID(state)));
  PCC("_data *data;\n");
  PCC("  if (!this) return 0;\n");
  PCC("  data = &this->u.");
  PCC(get_string(STATE_ID(state)));
  PCC(";\n");
  PCC("  if (this->_case_fun == NULL)\n    data->_case.");
  PCC(get_string(ACTION_ID(state, action)));
  PCC(" = ");
  sprintf(buf, "_m_%d_%d_0", state, action);
  PCC(buf);
  PCC(";\n");
  PCC("\n");
  PCC("  switch (data->_case.");
  PCC(get_string(ACTION_ID(state, action)));
  PCC(") {\n");

  /* the (generated) switch of death - if you want to hack cgscore
   * and if there is no external documentation, contact the authors,
   * directly reading the following code is not the way to get the
   * point (at the time of writing of this comment the mechanism is
   * not documented, it might (well it should) be done in the future)
   */
  for (i = 0; i < CODE_NB_CALLS(code) + 1; i++) {
    PCC("    case _m_");
    sprintf(buf, "%d_%d_%d", state, action, i);
    PCC(buf);
    /* barbarian use of ';' to be able to define variables at this point
     * C99 is weird.
     */
    PCC(":;\n");
    //PCC("    {\n");

    if (i != CODE_NB_CALLS(code))
      if (put_define_call(state, action, i) == -1) return -1;

    if (i == 0)
      if (put_va_arg(state, action) == -1) return -1;

    if (put_code(state, action, i) == -1) return -1;

    PCC("\n");

    //PCC("    }\n");
  }

  PCC("  }\n");
  PCC("  return 0;\n");
  PCC("}\n");

  return 0;
}

/* 0 ok, -1 error */
int put_actions_state(int s)
{
  int i;

  Vprintf(stderr, ".. entering %s\n", __FUNCTION__);

  for (i = 0; i < STATE_NB_ACTIONS(s); i++)
    if (put_action(s, i) == -1) return -1;

  return 0;
}

/* 0 ok, -1 error */
int put_actions(void)
{
  int i;
  char buf[32];

  Vprintf(stderr, ".. entering %s\n", __FUNCTION__);

  if (split)
    PC("\n");

  for (i = 0; i < list_size(states); i++) {
    if (split) {
      CC = get_string(STATE_FILENAME(i));
      cc = fopen(CC, "a"); if (!cc) { perror(CC); return -1; }
    }

    PCC("\n#define BREAK SUCCESS\n");

    if (put_actions_state(i) == -1) return -1;
    if (split) {
      if (fclose(cc) == -1) { perror(CC); return -1; }
    }
  }

  PC("\n");
  PC("struct actions actions[");
  sprintf(buf, "%d", list_size(states)+1);
  PC(buf);
  PC("] = {\n");
  PC("  /* _m_type_0_state */\n");
  PC("  {},\n");

  for (i = 0; i < list_size(states); i++)
    if (put_state_actions(i) == -1) return -1;

  PC("};\n");

  return 0;
}

/* 0 ok, -1 error */
int create_states(void)
{
  int i;

  Vprintf(stderr, ".. entering %s\n", __FUNCTION__);

  if (hout) H = hout;

  c = fopen(C, "w"); if (!c) { perror(C); return -1; }
  if (!only_c) { h = fopen(H, "w"); if (!h) { perror(H); return -1; } }
  if (!split) {
    cc = c;
    CC = C;
  }

  /* we destroy files if split mode (no check made) */
  if (split) {
    for (i = 0; i < list_size(states); i++)
      remove(get_string(STATE_FILENAME(i)));
  }

  /* header */
  PC("/* This file was generated by 'm'. ");
  PC("You probably should not modify it. */\n\n");

  PH("/* This file was generated by 'm'. ");
  PH("You probably should not modify it. */\n\n");
  PH("#ifndef _STATES_H_\n#define _STATES_H_\n\n");

  /* includes */
  PC("#include \"states.h\"\n");
  PC("#include \"error.h\"\n");
  PC("#include <stdlib.h>\n");
  PC("#include <stdarg.h>\n");
  PH("#include \"thread.h\"\n");
  PH("#include \"typedefs.h\"\n");

  for (i=0; i<list_size(include); i++) {
    PC("#include");
    PC(get_string((string)list_get_int(include, i)));
    PC("\n");

    PH("#include");
    PH(get_string((string)list_get_int(include, i)));
    PH("\n");
  }

  for (i=0; i<list_size(cinclude); i++) {
    PC("#include");
    PC(get_string((string)list_get_int(cinclude, i)));
    PC("\n");
  }

  for (i=0; i<list_size(hinclude); i++) {
    PH("#include");
    PH(get_string((string)list_get_int(hinclude, i)));
    PH("\n");
  }

  PC("\n");
  PH("\n");

  /* put the header and include in the CC files */
  if (!split) goto no_split;

  for (i = 0; i < list_size(states); i++) {
    int j;

    CC = get_string(STATE_FILENAME(i));
    cc = fopen(CC, "a"); if (!cc) { perror(CC); return -1; }

    /* header */
    PCC("/* This file was generated by 'm'. ");
    PCC("You probably should not modify it. */\n\n");

    /* includes */
    PCC("#include \"states.h\"\n");
    PCC("#include \"error.h\"\n");
    PCC("#include <stdlib.h>\n");
    PCC("#include <stdarg.h>\n");

    for (j=0; j<list_size(include); j++) {
      PCC("#include");
      PCC(get_string((string)list_get_int(include, j)));
      PCC("\n");
    }

    for (j=0; j<list_size(cinclude); j++) {
      PCC("#include");
      PCC(get_string((string)list_get_int(cinclude, j)));
      PCC("\n");
    }

    PCC("\n");

    if (fclose(cc) == -1) { perror(CC); return -1; }
  }

no_split:
  /* feed the .h */
  /* the structure actions */
  if (put_struct_actions() == -1) return -1;

  /* the action macros */
  if (put_action_macros() == -1) return -1;

  /* feed the .c */
  /* enum & structs definitions */
  if (put_enum_structs() == -1) return -1;

  /* the new_state function */
  if (split)
    PH("\nstruct _m_state *new_state(enum _m_state_type);\n");
  PC("/* return 0 if error, malloced memory if ok */\n");
  PC("struct _m_state *new_state(enum _m_state_type t)\n");
  PC("{\n");
  PC("  struct _m_state *ret = malloc(sizeof(struct _m_state));\n");
  PC("  if (!ret) return 0;\n");
  PC("  ret->next = 0;\n");
  PC("  ret->previous = 0;\n");
  PC("  ret->type = t;\n");
  PC("  return ret;\n");
  PC("}\n");

  /* enter functions */
  if (put_enters() == -1) return -1;

  /* leave functions */
  if (put_leaves() == -1) return -1;

  /* actions */
  if (put_actions() == -1) return -1;

  /* trailer */
  PH("\n");
  PH("#endif /* _STATES_H_ */\n");

  if (fclose(c) == -1) { perror(C); return -1; }
  if (!only_c) if (fclose(h) == -1) { perror(H); return -1; }

  return 0;
}
