/* gcomposer: a little tool to write music.
 * http://sed.free.fr/gcomposer
 * Released in the public domain (freedom has no license).
 */

#include "options.h"

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

void default_options(options *c)
{
  (void)strcpy(c->audio_oss_file, "/dev/dsp");
  (void)strcpy(c->audio_midi_file, "/dev/midi");
#ifndef ALSA
  c->ptype = PLAY_MIDI;
#else
  c->ptype = PLAY_MIDI_ALSA;
#endif
  c->midi_device = 0;
  c->midi_instrument = 0;
  c->midi_adjust = 0;
  c->midi_volume = 64;
}

#define SKIP						\
  /* go to '=' */					\
  while (line[pos] && line[pos] != '=') pos++;		\
  if (line[pos] != '=') continue;			\
  /* skip spaces after '=' */				\
  do pos ++; while (line[pos] && isspace(line[pos]));	\
  if (!line[pos]) continue

/* parsing the option's file accepts more stuff than it should,
 * but let's keep that simple, no need to be anal
 */
void load_options(options *c)
{
  char n[FILENAME_MAX];
  char line[FILENAME_MAX+256];
  int pos;
  char *t;
  FILE *f;

  default_options(c);

#ifndef WINDOZ
  if (!(t=getenv("HOME"))) return;
  if (strlen(t) + strlen("/.gcomposer") + 1 > FILENAME_MAX) return;

  sprintf(n, "%s/.gcomposer", t);
#else /* WINDOZ */
  /* stupid windows with their fucking \ shit */
  if (!(t=getenv("USERPROFILE"))) return;
  if (strlen(t) + strlen("\\\\.gcomposer") + 1 > FILENAME_MAX) return;

  sprintf(n, "%s\\\\.gcomposer", t);
#endif /* WINDOZ */

  f = fopen(n, "r");
  if (!f) return;

  while (1) {
    /* read a line */
    pos = 0;
    while (pos < FILENAME_MAX+255) {
      int c = getc(f);
      if (c == EOF) goto read_end;
      if (c == '\n') break;
      line[pos] = c;
      pos++;
    }
    line[pos] = 0;
    pos = 0;
    /* get the option */
    if (!strncmp(line, "oss", 3)) {
      pos += 3;
      SKIP;
      if (strlen(line+pos) > FILENAME_MAX-1 || line[pos] == 0) continue;
      (void)strcpy(c->audio_oss_file, line+pos);
    } else
    if (!strncmp(line, "midi_device", 11)) {
      pos += 11;
      SKIP;
      if (line[pos] == '[')
        c->midi_device = -1;
      else
        sscanf(line+pos, "%d", &c->midi_device);
    } else
    if (!strncmp(line, "midi_instrument", 15)) {
      pos += 15;
      SKIP;
      sscanf(line+pos, "%d", &c->midi_instrument);
    } else
    if (!strncmp(line, "midi_adjust", 11)) {
      pos += 11;
      SKIP;
      sscanf(line+pos, "%d", &c->midi_adjust);
    } else
    if (!strncmp(line, "midi_volume", 11)) {
      pos += 11;
      SKIP;
      sscanf(line+pos, "%d", &c->midi_volume);
      if (c->midi_volume < 0) c->midi_volume = 0;
      if (c->midi_volume > 127) c->midi_volume = 127;
    } else
    if (!strncmp(line, "midi", 4)) {
      pos += 4;
      SKIP;
      if (strlen(line+pos) > FILENAME_MAX-1 || line[pos] == 0) continue;
      (void)strcpy(c->audio_midi_file, line+pos);
    } else
    if (!strncmp(line, "type", 4)) {
      pos += 4;
      SKIP;
      if (line[pos] == 'o')
        c->ptype = PLAY_OSS;
      else if (line[pos] == 'm')
        c->ptype = PLAY_MIDI;
      else if (line[pos] == 'a')
#ifdef ALSA
        c->ptype = PLAY_MIDI_ALSA;
#else
        c->ptype = PLAY_MIDI;
#endif
    }
  }

read_end:
  fclose(f);
}

/* not very robust function, no checks and so on... */
/* I AM A FUCKING BAD HACKER! */
void save_options(options *c)
{
  char n[FILENAME_MAX];
  char *t;
  FILE *f;

#ifndef WINDOZ
  if (!(t=getenv("HOME"))) return;
  if (strlen(t) + strlen("/.gcomposer") + 1 > FILENAME_MAX) return;

  sprintf(n, "%s/.gcomposer", t);
#else /* WINDOZ */
  /* stupid windows with their fucking \ shit */
  if (!(t=getenv("USERPROFILE"))) return;
  if (strlen(t) + strlen("\\\\.gcomposer") + 1 > FILENAME_MAX) return;

  sprintf(n, "%s\\\\.gcomposer", t);
#endif /* WINDOZ */

  f = fopen(n, "w");
  if (!f) return;

  fprintf(f, "oss=%s\n", c->audio_oss_file);
  fprintf(f, "midi=%s\n", c->audio_midi_file);
  fprintf(f, "type=%s\n",
          c->ptype == PLAY_OSS ? "oss" :
              c->ptype == PLAY_MIDI_ALSA ? "alsa" : "midi");
  if (c->midi_device == -1)
    fprintf(f, "midi_device=[none]\n");
  else
    fprintf(f, "midi_device=%d\n", c->midi_device);
  fprintf(f, "midi_instrument=%d\n", c->midi_instrument);
  fprintf(f, "midi_adjust=%d\n", c->midi_adjust);
  fprintf(f, "midi_volume=%d\n", c->midi_volume);
  fclose(f);
}
