/* gcomposer: a little tool to write music.
 * http://sed.free.fr/gcomposer
 * Released in the public domain (freedom has no license).
 */

action about {
  CALL(about_state, &data->d->cursor);
}

action authors {
  CALL(authors_state, &data->d->cursor);
}

action options {
  CALL(options_state, data->d);
}

action new {
  if (data->d->state == STATE_MODIFIED) {
    CALL(confirm_new_state, data->ret_code, data->g->win);
    if (*data->ret_code == 0) BREAK;
  }
  destroy_doc(data->d);
  data->d->cursor.type = BIG_CURSOR;
  data->d->cursor.line = 0;
  data->d->cursor.chord = 0;
  data->d->tempo = 80;
  data->d->sel.start_line = -1;
  data->d->state = STATE_NEW;
  if (!doc_insert_line(data->d, LINE_MUSIC, -1)) {
    /* we are in a big shit here, let's just quit */
    /* in normal situations we should never come here */
    set_error(FATAL);
    CALL(error_state, data->g->win);
    exit(1);
  }
  size_reset(data->d);
  size_post_treatment(data->d);
  reset_cursor(&data->d->cursor);
  set_info(data->d, "");
  display_tempo(data->d);
  if (this->type != _m_type_init_state)
    RETURN;
}

action save {
  /* is there a filename? */
  if (data->d->filename[0] == 0) {
    CALL(save_state, data->d->filename, data->g->win, &data->d->cursor);
    if (data->d->filename[0] == 0) BREAK;
  }
  if (!save(data->d, data->d->filename)) {
    CALL(error_state, data->g->win);
    BREAK;
  }
  data->d->state = STATE_SAVED;
  set_info(data->d, "File %s succesfully saved", data->d->filename);
}

action save_as {
  CALL(save_as_state, data->d->filename, data->g->win, &data->d->cursor);
  if (data->d->filename[0] == 0) BREAK;
  if (!save(data->d, data->d->filename)) {
    data->d->filename[0] = 0;
    CALL(error_state, data->g->win);
    BREAK;
  }
  data->d->state = STATE_SAVED;
  set_info(data->d, "File %s succesfully saved", data->d->filename);
}

action load {
  CALL(load_state, data->load_file, data->g->win, &data->d->cursor);
  if (data->load_file[0] == 0) BREAK;
  if (data->d->state == STATE_MODIFIED) {
    CALL(confirm_load_state, data->ret_code, data->g->win);
    if (*data->ret_code == 0) BREAK;
  }
  if (!load(data->d, data->load_file)) {
    CALL(error_state, data->g->win);
    BREAK;
  }
  strcpy(data->d->filename, data->load_file);
  data->d->state = STATE_SAVED;
  size_load(data->d);
  size_post_treatment(data->d);
  set_info_chord(data->d);
  display_tempo(data->d);
  if (this->type != _m_type_init_state)
    RETURN;
  DO_GOTO_STATE(_owner);
}

action quit {
  if (data->d->state == STATE_MODIFIED) {
    CALL(confirm_quit_state, data->ret_code, data->g->win);
    if (*data->ret_code == 0) BREAK;
  }
  *data->ret_code = I_QUIT;
  RETURN;
}

action undo {
  dprintf("undo\n");
  clear_cursor(&data->d->cursor);
  data->d->cursor.type = BIG_CURSOR;
  history_up(data->d);
  size_post_treatment(data->d);
  reset_cursor(&data->d->cursor);
  set_info_chord(data->d);

  /* set the undo/redo grey stuff */
  set_grey_undo_redo(data->d);

  if (this->type != _m_type_init_state) {
    *data->ret_code = I_UNDO_REDO;
    dprintf("return from undo\n");
    RETURN;
  }
  DO_GOTO_STATE(_owner);
}

action redo {
  clear_cursor(&data->d->cursor);
  data->d->cursor.type = BIG_CURSOR;
  history_down(data->d, 0);
  size_post_treatment(data->d);
  reset_cursor(&data->d->cursor);
  set_info_chord(data->d);

  /* set the undo/redo grey stuff */
  set_grey_undo_redo(data->d);

  if (this->type != _m_type_init_state) {
    *data->ret_code = I_UNDO_REDO;
    RETURN;
  }
  DO_GOTO_STATE(_owner);
}

action tutorial {
  disable_cursor(&data->d->cursor);
  CALL(tutorial_init, data->g);
  enable_cursor(&data->d->cursor);
  reset_cursor(&data->d->cursor);
}

action selection_paste {
  /* no paste while in select mode */
  if (this->type == _m_type_state_selection) BREAK;
  history_next_is_major(&data->d->history);
  clear_cursor(&data->d->cursor);
  data->d->cursor.type = BIG_CURSOR;
  if (!paste_selection(data->d)) {
    history_up_and_forget(data->d);
    size_post_treatment(data->d);
    CALL(error_state, data->g->win);
    goto end_paste;
  }
  size_post_treatment(data->d);
end_paste:
  reset_cursor(&data->d->cursor);
  set_info_chord(data->d);
  if (this->type != _m_type_init_state) {
    *data->ret_code = I_PASTE;
    RETURN;
  }
  DO_GOTO_STATE(_owner);
}

action mouse_click(int x, int y)
{
  clear_cursor(&data->d->cursor);
  position_cursor(&data->d->cursor, x / data->d->zoom, y / data->d->zoom);
  reset_cursor(&data->d->cursor);
  set_info_chord(data->d);
  if (this->type != _m_type_init_state) {
    *data->ret_code = I_NOTHING;
    RETURN;
  }
  DO_GOTO_STATE(_owner);
}
