/* gcomposer: a little tool to write music.
 * http://sed.free.fr/gcomposer
 * Released in the public domain (freedom has no license).
 */

state i_state(doc *d, gui *g, int *ret_code)
{
  data {
    doc *d;
    struct gui *g;
    int *ret_code;
    char load_file[FILENAME_MAX];
  }

  enter {
    dprintf("enter i state\n");
    data->d = d;
    data->g = g;
    data->ret_code = ret_code;
    clear_cursor(&data->d->cursor);
    data->d->cursor.type = SMALL_CURSOR;
    data->d->cursor.note = 0;
    reset_cursor(&data->d->cursor);
    *data->ret_code = I_NOTHING;
  }

  leave {
    dprintf("leave i state\n");
    clear_cursor(&data->d->cursor);
    data->d->cursor.type = BIG_CURSOR;
    reset_cursor(&data->d->cursor);
  }

  action key_press(int key, int control, int shift) {
    int pitch_direction;

    if (control) {
      switch (key) {
        case 's':
          DO_SAVE(_owner);
          /* after a DO_XX, we must return 0 */
          return 0;
        case 'z':
          DO_UNDO(_owner);
          return 0;
        case 'y':
          DO_REDO(_owner);
          return 0;
        case 'l': pitch_direction = -1; goto case_head_delta;
        case 'L': pitch_direction = 1; goto case_head_delta;
        case 'm': pitch_direction = -1; goto case_alteration_delta;
        case 'M': pitch_direction = 1; goto case_alteration_delta;
        case 'o': pitch_direction = -1; goto case_dots_delta;
        case 'O': pitch_direction = 1; goto case_dots_delta;
      }
      goto end_switch;
    }

    switch (key) {
      case GDK_Escape: goto case_return;
      case GDK_Up: goto case_up;
      case GDK_Down: goto case_down;
      case GDK_Delete: goto case_delete;
      case 't': pitch_direction = 1; goto case_time;
      case 'T': pitch_direction = -1; goto case_time;
      case 'u': pitch_direction = 1; goto case_pitch;
      case 'U': pitch_direction = -1; goto case_pitch;
      case '0': case '1': case '2': case '3': case '4':
      case '5': case '6': case '7': case '8': case '9':
        goto case_octave;
      case '#': goto case_diese;
      case '.': goto case_point;
      case 'i': goto case_return;
      case 'a': case 'b': case 'c': case 'd':
      case 'e': case 'f': case 'g':
        goto case_insert_note;
      case 'A': case 'B': case 'C': case 'D':
      case 'F': case 'E': case 'G':
        key += 'a' - 'A';
        goto case_insert_note;
      case 's': case 'S': goto case_insert_silence;
      case 'r': goto case_redraw;
      case ' ':
      case 'p': goto case_start_play2;
      case 'l': pitch_direction = -1; goto case_head_delta;
      case 'L': pitch_direction = 1; goto case_head_delta;
      case 'm': pitch_direction = -1; goto case_alteration_delta;
      case 'M': pitch_direction = 1; goto case_alteration_delta;
      case 'o': pitch_direction = -1; goto case_dots_delta;
      case 'O': pitch_direction = 1; goto case_dots_delta;
      case 'j': pitch_direction = -1; goto case_stem;
      case 'J': pitch_direction = 1; goto case_stem;
      default: goto end_switch;
    }

case_up:
    if (data->d->cursor.note != CHORD_SIZE(data->d, data->d->cursor.line,
                                           data->d->cursor.chord)-1) {
      clear_cursor(&data->d->cursor);
      data->d->cursor.note++;
      set_info_chord(data->d);
    }
    goto end_switch;

case_down:
    if (data->d->cursor.note != 0) {
      clear_cursor(&data->d->cursor);
      data->d->cursor.note--;
      set_info_chord(data->d);
    }
    goto end_switch;

case_delete:
    /* if there is only one note, we also delete the chord */
    if (CHORD_SIZE(data->d, data->d->cursor.line, data->d->cursor.chord)
          == 1) {
      history_next_is_major(&data->d->history);
      clear_cursor(&data->d->cursor);
      data->d->cursor.type = BIG_CURSOR;
      if (!delete_chord(data->d, data->d->cursor.line,
                        data->d->cursor.chord)) {
        history_up_and_forget(data->d);
        size_post_treatment(data->d);
        CALL(error_state, data->g->win);
        goto end_switch;
      }
      size_post_treatment(data->d);
      set_info_chord(data->d);
      RETURN;
    }
    /* more than one note, delete only it */
    history_next_is_major(&data->d->history);
    clear_cursor(&data->d->cursor);
    if (!delete_note(data->d, data->d->cursor.line, data->d->cursor.chord,
                     data->d->cursor.note)) {
      history_up_and_forget(data->d);
      size_post_treatment(data->d);
      CALL(error_state, data->g->win);
      goto end_switch;
    }
    size_post_treatment(data->d);
    /* if we deleted the last note of the chord, point to the prev */
    if (CHORD_SIZE(data->d, data->d->cursor.line, data->d->cursor.chord)
          == data->d->cursor.note) {
      data->d->cursor.note--;
    }
    set_info_chord(data->d);
    goto end_switch;

case_time: {
    int duration;

    switch (NOTE_TYPE(data->d, data->d->cursor.line,
                      data->d->cursor.chord, data->d->cursor.note)) {
      case NOTE:
        duration = NOTE_DURATION(data->d, data->d->cursor.line,
                                 data->d->cursor.chord, data->d->cursor.note);
        break;
      case SILENCE:
        duration = SILENCE_DURATION(data->d, data->d->cursor.line,
                                 data->d->cursor.chord, data->d->cursor.note);
        break;
    }
    switch (duration) {
      case RONDE:
        if (pitch_direction == 1) duration = BLANCHE;
        break;
      case BLANCHE:
        if (pitch_direction == 1) duration = NOIRE; else duration = RONDE;
        break;
      case NOIRE:
        if (pitch_direction == 1) duration = CROCHE; else duration = BLANCHE;
        break;
      case CROCHE:
        if (pitch_direction == 1) duration = DCROCHE; else duration = NOIRE;
        break;
      case DCROCHE:
        if (pitch_direction == 1) duration = DDCROCHE; else duration = CROCHE;
        break;
      case DDCROCHE:
        if (pitch_direction == 1) duration = DDDCROCHE;
        else duration = DCROCHE;
        break;
      case DDDCROCHE:
        if (pitch_direction == 1) duration = DDDDCROCHE;
        else duration = DDCROCHE;
        break;
      case DDDDCROCHE:
        if (pitch_direction == -1) duration = DDDCROCHE;
        break;
    }
    history_next_is_major(&data->d->history);
    clear_cursor(&data->d->cursor);
    if (set_note_duration(data->d, data->d->cursor.line,
                          data->d->cursor.chord,
                          data->d->cursor.note, duration) < 0) {
      history_up_and_forget(data->d);
      size_post_treatment(data->d);
      goto set_error;
    }
    size_post_treatment(data->d);
    set_info_chord(data->d);
    goto end_switch;
  }

case_pitch: {
    int pitch;
    int n;
    switch (NOTE_TYPE(data->d, data->d->cursor.line,
                      data->d->cursor.chord, data->d->cursor.note)) {
      case NOTE:
        pitch = NOTE_PITCH(data->d, data->d->cursor.line,
                           data->d->cursor.chord, data->d->cursor.note);
        pitch += pitch_direction;
        history_next_is_major(&data->d->history);
        clear_cursor(&data->d->cursor);
        if (!(n = set_note_pitch(data->d, data->d->cursor.line,
                                 data->d->cursor.chord,
                                 data->d->cursor.note, pitch)))
          goto end_switch;
        if (n < 0) {
          history_up_and_forget(data->d);
          size_post_treatment(data->d);
          goto set_error;
        }
        size_post_treatment(data->d);
        break;
      case SILENCE:
        pitch = SILENCE_POSITION(data->d, data->d->cursor.line,
                                 data->d->cursor.chord, data->d->cursor.note);
        pitch += pitch_direction;
        history_next_is_major(&data->d->history);
        clear_cursor(&data->d->cursor);
        if (!(n = set_note_position(data->d, data->d->cursor.line,
                                    data->d->cursor.chord,
                                    data->d->cursor.note, pitch)))
          goto end_switch;
        if (n < 0) {
          history_up_and_forget(data->d);
          size_post_treatment(data->d);
          goto set_error;
        }
        size_post_treatment(data->d);
        break;
    }
    data->d->cursor.note = n - 1;
    set_info_chord(data->d);
    goto end_switch;
  }

case_octave: {
    int octave = key - '0';
    int n;
    switch (NOTE_TYPE(data->d, data->d->cursor.line,
                      data->d->cursor.chord, data->d->cursor.note)) {
      case NOTE: {
        int pitch = NOTE_PITCH(data->d, data->d->cursor.line,
                               data->d->cursor.chord, data->d->cursor.note);
        int delta_clef = get_delta_clef(data->d, data->d->cursor.line,
                                        data->d->cursor.chord);
        pitch -= delta_clef;
        /* redo, too slow (but hey, it won't loop 1,000,000 times...) */
        /* the 4th octave is from C4 to B4, thus the -4 and 3 below */
        while (pitch < -4) pitch += 7;
        while (pitch >= 3) pitch -= 7;
        pitch += 7 * (octave - 4);
        pitch += delta_clef;
        /* pitch not changed? */
        if (pitch == NOTE_PITCH(data->d, data->d->cursor.line,
                                data->d->cursor.chord, data->d->cursor.note))
          goto end_switch;
        history_next_is_major(&data->d->history);
        clear_cursor(&data->d->cursor);
        if (!(n = set_note_pitch(data->d, data->d->cursor.line,
                                 data->d->cursor.chord,
                                 data->d->cursor.note, pitch)))
          goto end_switch;
        if (n < 0) {
          history_up_and_forget(data->d);
          size_post_treatment(data->d);
          goto set_error;
        }
        size_post_treatment(data->d);
        break;
      }
      case SILENCE: /* do nothing */; break;
    }
    data->d->cursor.note = n - 1;
    set_info_chord(data->d);
    goto end_switch;
  }

case_diese: {
    int alteration = NOTE_ALTERATION(data->d, data->d->cursor.line,
                           data->d->cursor.chord, data->d->cursor.note);
    switch (alteration) {
      case NORMAL: alteration = DIESE; break;
      case DIESE: alteration = BEMOL; break;
      case BEMOL: alteration = BECARRE; break;
      case BECARRE: alteration = NORMAL; break;
    }
    history_next_is_major(&data->d->history);
    clear_cursor(&data->d->cursor);
    if (set_note_alteration(data->d, data->d->cursor.line,
                            data->d->cursor.chord,
                            data->d->cursor.note, alteration) < 0) {
      history_up_and_forget(data->d);
      size_post_treatment(data->d);
      goto set_error;
    }
    size_post_treatment(data->d);
    set_info_chord(data->d);
    goto end_switch;
  }

case_point: {
    int nb_dots;

    switch (NOTE_TYPE(data->d, data->d->cursor.line,
                      data->d->cursor.chord, data->d->cursor.note)) {
      case NOTE:
        nb_dots = NOTE_NB_DOTS(data->d, data->d->cursor.line,
                               data->d->cursor.chord, data->d->cursor.note);
        break;
      case SILENCE:
        nb_dots = SILENCE_NB_DOTS(data->d, data->d->cursor.line,
                               data->d->cursor.chord, data->d->cursor.note);
        break;
    }
    nb_dots++;
    if (nb_dots == 4) nb_dots = 0;
    history_next_is_major(&data->d->history);
    clear_cursor(&data->d->cursor);
    if (set_note_dot(data->d, data->d->cursor.line,
                     data->d->cursor.chord,
                     data->d->cursor.note, nb_dots) < 0) {
      history_up_and_forget(data->d);
      size_post_treatment(data->d);
      goto set_error;
    }
    size_post_treatment(data->d);
    set_info_chord(data->d);
    goto end_switch;
  }

case_return:
    RETURN;

case_insert_note: {
    int n;
    switch (key) {
      case 'a': n = 1; break;
      case 'b': n = 2; break;
      case 'c': n = 3; break;
      case 'd': n = 4; break;
      case 'e': n = 5; break;
      case 'f': n = -1; break;
      case 'g': n = 0; break;
    }
    n += get_delta_clef(data->d, data->d->cursor.line,
                        data->d->cursor.chord);
    history_next_is_major(&data->d->history);
    clear_cursor(&data->d->cursor);
    if (!(n = insert_note(data->d, data->d->cursor.line,
                          data->d->cursor.chord,
                          NOTE, n, NORMAL, get_gui_duration(data->g),
                          0, 0, 0, 0, 0))) {
      history_up_and_forget(data->d);
      size_post_treatment(data->d);
      CALL(error_state, data->g->win);
      goto end_switch;
    }
    size_post_treatment(data->d);
    data->d->cursor.note = n - 1;
    set_info_chord(data->d);
    goto end_switch;
  }

case_insert_silence: {
    int n;
    history_next_is_major(&data->d->history);
    clear_cursor(&data->d->cursor);
    if (!(n = insert_note(data->d, data->d->cursor.line,
                          data->d->cursor.chord,
                          SILENCE, 0, get_gui_duration(data->g), 0))) {
      history_up_and_forget(data->d);
      size_post_treatment(data->d);
      CALL(error_state, data->g->win);
      goto end_switch;
    }
    size_post_treatment(data->d);
    data->d->cursor.note = n - 1;
    set_info_chord(data->d);
    goto end_switch;
  }

case_head_delta: {
    if (NOTE_TYPE(data->d, data->d->cursor.line,
                  data->d->cursor.chord, data->d->cursor.note) != NOTE)
      goto end_switch;
    if (!control)
      pitch_direction *= images[I_BLACK_HEAD].width-1;
    pitch_direction += NOTE_HEAD_DELTA(data->d,
                                       data->d->cursor.line,
                                       data->d->cursor.chord,
                                       data->d->cursor.note);
    history_next_is_major(&data->d->history);
    clear_cursor(&data->d->cursor);
    if (set_note_head_delta(data->d, data->d->cursor.line,
                            data->d->cursor.chord,
                            data->d->cursor.note, pitch_direction) < 0) {
      history_up_and_forget(data->d);
      size_post_treatment(data->d);
      goto set_error;
    }
    size_post_treatment(data->d);
    set_info_chord(data->d);
    goto end_switch;
  }

case_alteration_delta: {
    if (NOTE_TYPE(data->d, data->d->cursor.line,
                  data->d->cursor.chord, data->d->cursor.note) != NOTE)
      goto end_switch;
    if (!control)
      pitch_direction *= images[I_BLACK_HEAD].width-1;
    pitch_direction += NOTE_ALTERATION_DELTA(data->d,
                                             data->d->cursor.line,
                                             data->d->cursor.chord,
                                             data->d->cursor.note);
    history_next_is_major(&data->d->history);
    clear_cursor(&data->d->cursor);
    if (set_note_alteration_delta(data->d, data->d->cursor.line,
                                  data->d->cursor.chord,
                                  data->d->cursor.note, pitch_direction) < 0) {
      history_up_and_forget(data->d);
      size_post_treatment(data->d);
      goto set_error;
    }
    size_post_treatment(data->d);
    set_info_chord(data->d);
    goto end_switch;
  }

case_dots_delta: {
    if (NOTE_TYPE(data->d, data->d->cursor.line,
                  data->d->cursor.chord, data->d->cursor.note) != NOTE)
      goto end_switch;
    if (!control)
      pitch_direction *= images[I_BLACK_HEAD].width-1;
    pitch_direction += NOTE_DOTS_DELTA(data->d,
                                       data->d->cursor.line,
                                       data->d->cursor.chord,
                                       data->d->cursor.note);
    history_next_is_major(&data->d->history);
    clear_cursor(&data->d->cursor);
    if (set_note_dots_delta(data->d, data->d->cursor.line,
                            data->d->cursor.chord,
                            data->d->cursor.note, pitch_direction) < 0) {
      history_up_and_forget(data->d);
      size_post_treatment(data->d);
      goto set_error;
    }
    size_post_treatment(data->d);
    set_info_chord(data->d);
    goto end_switch;
  }

case_stem: {
    if (NOTE_TYPE(data->d, data->d->cursor.line,
                  data->d->cursor.chord, data->d->cursor.note) != NOTE)
      goto end_switch;
    history_next_is_major(&data->d->history);
    clear_cursor(&data->d->cursor);
    if (set_note_stem_down(data->d, data->d->cursor.line,
                           data->d->cursor.chord,
                           data->d->cursor.note, pitch_direction == -1) < 0) {
      history_up_and_forget(data->d);
      size_post_treatment(data->d);
      goto set_error;
    }
    size_post_treatment(data->d);
    set_info_chord(data->d);
    goto end_switch;
  }

case_redraw:
    size_reset(data->d);
    size_post_treatment(data->d);
    invalidate_all(data->d);
    set_info_chord(data->d);
    goto end_switch;

case_start_play2:
    setup_tempo(data->d);
    if (!start_play(data->d, data->d->cursor.line, data->d->cursor.chord,
                    data->d->tempo)) {
      CALL(error_state, data->g->win);
      goto end_switch;
    }
    CALL(play_state, data->d);
    set_info_chord(data->d);
    RETURN;

set_error:
    CALL(error_state, data->g->win);
    goto end_switch;

end_switch:
    reset_cursor(&data->d->cursor);

    /* set the grey for undo/redo */
    set_grey_undo_redo(data->d);
  }

%%import "states/common_actions.i"
}
