/* gcomposer: a little tool to write music.
 * http://sed.free.fr/gcomposer
 * Released in the public domain (freedom has no license).
 */

state init_state
{
  data {
    struct gui _g;
    doc _d;
    /* pointers to use common_actions.i like the other states */
    struct gui *g;
    doc *d;
    char load_file[FILENAME_MAX];
    int ret;
    int key;
    int control;
    int shift;
    /* unused but needed to work with common_actions.i */
    int *ret_code;
  }

  enter {
    parachute_doc = &data->_d;
    data->d = &data->_d;
    data->g = &data->_g;
    data->ret_code = &data->ret;
    memset(data->d, 0, sizeof(doc));
    init_gui(data->g, this->thread, data->d);
    gtk_wrapper_widget_show_all(data->g->win);
    gtk_wrapper_create_gc(data->g->win);
    data->d->drawing_area = data->g->d;
    data->d->tempo_text = data->g->tempo_entry;
    data->d->fonts = &data->g->fonts[0];
    data->d->desc_label = data->g->desc;

    data->d->cursor.doc = data->d;
    data->d->cursor.type = BIG_CURSOR;
    data->d->cursor.line = 0;
    data->d->cursor.chord = 0;
    data->d->cursor.hidden_level = 1;

    data->d->state = STATE_NEW;

    data->d->p.current_chord = -1;
    data->d->tempo = 80;

    data->d->sel.start_line = -1;

    data->d->current_width = -1;
    data->d->current_height = -1;

    data->d->zoom = 1;

    load_options(&data->d->options);
    init_midi(&data->d->midi);
#ifdef ALSA
    if (data->d->options.ptype == PLAY_MIDI_ALSA)
      data->d->midi.is_alsa = 1;
#endif

    INIT_HISTORY(data->d->history);

    /* let's be barbarian with this first insert */
    /* if it fails nothing can be done at all */
    /* we are barbarian because an 'enter' is not supposed to fail,
     * so we have no 'clean' way to handle this case (yeah, a
     * rewrite of 'm' is required, but the heck, here doc_insert_line
     * will *NEVER* fail, so not a big issue)
     */
    if (!doc_insert_line(data->d, LINE_MUSIC, -1)) exit(1);

    size_reset(data->d);
    size_post_treatment(data->d);
    reset_cursor(&data->d->cursor);

    /* allow gui elements */
    ungrey(GREY_MENU_FILE);
    ungrey(GREY_MENU_EDIT);
    ungrey(GREY_MENU_TOOL);
    ungrey(GREY_MENU_HELP);
    ungrey(GREY_MENU_QUIT);
    ungrey(GREY_MENU_ZOOM);
    ungrey(GREY_MENU_NEW);
    ungrey(GREY_MENU_OPEN);
    ungrey(GREY_MENU_SAVE);
    ungrey(GREY_MENU_SAVE_AS);
    ungrey(GREY_MENU_OPTIONS);
    ungrey(GREY_MENU_VKEYBOARD);
    ungrey(GREY_MENU_ABOUT);
    ungrey(GREY_MENU_AUTHORS);
    ungrey(GREY_MENU_QUICK_HELP);
    ungrey(GREY_MENU_TUTORIAL);
    ungrey(GREY_MENU_KNOWN_BUGS);
    ungrey(GREY_MENU_TODO);
  }

  leave {
    close_cursor(&data->d->cursor);
    gtk_wrapper_widget_destroy(data->g->win);
    gtk_wrapper_main_quit();
    destroy_selection(data->d);
  }

  action key_press(int key, int control, int shift) {
    if (control) {
      switch (key) {
        case 'v': goto case_paste_selection;
        case 'z':
          DO_UNDO(_owner);
          /* after a DO_XX, we must return 0 */
          return 0;
        case 'y':
          DO_REDO(_owner);
          return 0;
        case 's':
          DO_SAVE(_owner);
          return 0;
      }
      goto end_switch;
    }

    /* we cannot use CALL (xx) into switch statement (because
     * 'm' creates another switch and labels will mix up)
     * so we do it this way
     */
    switch (key) {
#ifdef DEBUG_ON
      case 'h': {
        extern void dump_history(history *, int, int);
        if (data->d->history.size) dump_history(&data->d->history, 0, 2);
      }
#endif
      case GDK_Right:
        /* shift + right = enter selection mode (if we can go right) */
        if (shift) {
          if (data->d->cursor.chord ==
              data->d->data[data->d->cursor.line].u.music.size)
            goto end_switch;
          goto case_goto_selection;
        }
        goto case_GDK_Right;
      case GDK_Left:
        /* shift + left = enter selection mode (if we can go left) */
        if (shift) {
          if (data->d->cursor.chord == 0)
            goto end_switch;
          goto case_goto_selection;
        }
        goto case_GDK_Left;
      case GDK_Down:
        /* shift + down = enter selection mode (if we can go down) */
        if (shift) {
          if (data->d->cursor.line == data->d->size - 1)
            goto end_switch;
          goto case_goto_selection;
        }
        goto case_GDK_Down;
      case GDK_Up:
        /* shift + up = enter selection mode (if we can go up) */
        if (shift) {
          if (data->d->cursor.line == 0)
            goto end_switch;
          goto case_goto_selection;
        }
        goto case_GDK_Up;
      case GDK_Page_Up: if (!shift) goto case_page_up;
      case GDK_Page_Down: if (!shift) goto case_page_down;
      case GDK_End: goto case_GDK_End;
      case GDK_Home: goto case_GDK_Home;
      case GDK_Return:
        /* shift+return to insert a text line */
        if (shift)
          goto case_insert_text;
        goto case_GDK_Return;
      case GDK_BackSpace: goto case_GDK_BackSpace;
      case GDK_Delete: goto case_GDK_Delete;
      case GDK_Insert: if (shift) goto case_paste_selection; goto end_switch;
      case 'r': goto case_r;
      case 'a': case 'b': case 'c': case 'd':
      case 'e': case 'f': case 'g': goto case_insert_note;
      case 'A': case 'B': case 'C': case 'D':
      case 'F': case 'E': case 'G': key += 'a' - 'A'; goto case_insert_note;
      case 's': goto case_insert_silence;
      case 'S': goto case_insert_silence;
      case '*': goto case_tonalite;
      case '|': goto case_insert_break;
      case '&': goto case_insert_clef;
      case 'i': goto case_i;
      case ' ':
      case 'p': goto case_play;
      default: goto end_switch;
    }

case_GDK_Right:
    if (data->d->cursor.chord ==
        data->d->data[data->d->cursor.line].u.music.size)
      goto end_switch;
    clear_cursor(&data->d->cursor);
    data->d->cursor.chord++;
    set_info_chord(data->d);
    goto end_switch;

case_GDK_Left:
    if (data->d->cursor.chord == 0)
      goto end_switch;
    clear_cursor(&data->d->cursor);
    data->d->cursor.chord--;
    set_info_chord(data->d);
    goto end_switch;

case_GDK_End:
    clear_cursor(&data->d->cursor);
    data->d->cursor.chord =
        data->d->data[data->d->cursor.line].u.music.size;
    set_info_chord(data->d);
    goto end_switch;

case_GDK_Home:
    clear_cursor(&data->d->cursor);
    data->d->cursor.chord = 0;
    set_info_chord(data->d);
    goto end_switch;

case_GDK_Down:
    if (data->d->cursor.line == data->d->size - 1)
      goto end_switch;
    clear_cursor(&data->d->cursor);
    data->d->cursor.chord=0;
    data->d->cursor.line++;
    set_info_chord(data->d);
    goto end_switch;

case_GDK_Up:
    if (data->d->cursor.line == 0)
      goto end_switch;
    clear_cursor(&data->d->cursor);
    data->d->cursor.chord=0;
    data->d->cursor.line--;
    set_info_chord(data->d);
    goto end_switch;

case_page_down: {
    int height = gtk_wrapper_get_visible_height(data->d->drawing_area,
                                                data->d->zoom);
    clear_cursor(&data->d->cursor);
    data->d->cursor.chord=0;
    ensure_height_in_view(data->d,
        data->d->data[data->d->cursor.line].dim.y + height);
    while (height > 0) {
      if (data->d->cursor.line == data->d->size - 1) break;
      height -= data->d->data[data->d->cursor.line].dim.height;
      data->d->cursor.line++;
    }
    set_info_chord(data->d);
    goto end_switch;
  }

case_page_up: {
    int height = gtk_wrapper_get_visible_height(data->d->drawing_area,
                                                data->d->zoom);
    clear_cursor(&data->d->cursor);
    data->d->cursor.chord=0;
    while (height > 0) {
      if (!data->d->cursor.line) break;
      data->d->cursor.line--;
      height -= data->d->data[data->d->cursor.line].dim.height;
    }
    set_info_chord(data->d);
    goto end_switch;
  }

case_GDK_Return:
    history_next_is_major(&data->d->history);
    clear_cursor(&data->d->cursor);
    if (!split_line(data->d, data->d->cursor.line, data->d->cursor.chord)) {
      history_up_and_forget(data->d);
      size_post_treatment(data->d);
      CALL(error_state, data->g->win);
      goto end_switch;
    }
    size_post_treatment(data->d);
    data->d->cursor.chord=0;
    data->d->cursor.line++;
    goto end_switch;

case_GDK_BackSpace: {
    /* if beginning of line, merge if previous line is music line */
    if (data->d->cursor.chord == 0) {
      int prev_y, prev_width, prev_height;
      int chord_pos;
      if (data->d->cursor.line == 0)
        goto end_switch;
      if (LINE_TYPE(data->d, data->d->cursor.line-1) != LINE_MUSIC)
        goto end_switch;
      prev_y = data->d->data[data->d->cursor.line].dim.y;
      prev_width = data->d->data[data->d->cursor.line].dim.width;
      prev_height = data->d->data[data->d->cursor.line].dim.height;
      chord_pos = data->d->data[data->d->cursor.line-1].u.music.size;
      history_next_is_major(&data->d->history);
      clear_cursor(&data->d->cursor);
      if (!merge_lines(data->d, data->d->cursor.line-1)) {
        history_up_and_forget(data->d);
        size_post_treatment(data->d);
        CALL(error_state, data->g->win);
        goto end_switch;
      }
      size_post_treatment(data->d);
      data->d->cursor.chord = chord_pos;
      data->d->cursor.line--;
      goto end_switch;
    }
    /* otherwise delete chord */
    history_next_is_major(&data->d->history);
    clear_cursor(&data->d->cursor);
    if (!delete_chord(data->d, data->d->cursor.line,
                      data->d->cursor.chord-1)) {
      history_up_and_forget(data->d);
      size_post_treatment(data->d);
      CALL(error_state, data->g->win);
      goto end_switch;
    }
    size_post_treatment(data->d);
    data->d->cursor.chord--;
    goto end_switch;
  }

case_GDK_Delete:
    /* if end of line, merge if next line is music line */
    if (data->d->cursor.chord ==
          data->d->data[data->d->cursor.line].u.music.size) {
      int prev_y, prev_width, prev_height;
      if (data->d->cursor.line == data->d->size-1)
        goto end_switch;
      if (LINE_TYPE(data->d, data->d->cursor.line+1) != LINE_MUSIC)
        goto end_switch;
      prev_y = data->d->data[data->d->cursor.line].dim.y;
      prev_width = data->d->data[data->d->cursor.line].dim.width;
      prev_height = data->d->data[data->d->cursor.line].dim.height;
      history_next_is_major(&data->d->history);
      clear_cursor(&data->d->cursor);
      if (!merge_lines(data->d, data->d->cursor.line)) {
        history_up_and_forget(data->d);
        size_post_treatment(data->d);
        CALL(error_state, data->g->win);
        goto end_switch;
      } 
      size_post_treatment(data->d);
      set_info_chord(data->d);
      goto end_switch;
    }
    /* otherwise delete chord */
    history_next_is_major(&data->d->history);
    clear_cursor(&data->d->cursor);
    if (!delete_chord(data->d, data->d->cursor.line, data->d->cursor.chord)) {
      history_up_and_forget(data->d);
      size_post_treatment(data->d);
      CALL(error_state, data->g->win);
      goto end_switch;
    }
    size_post_treatment(data->d);
    set_info_chord(data->d);
    goto end_switch;

case_r:
    /* redraw */
    size_reset(data->d);
    size_post_treatment(data->d);
    invalidate_all(data->d);
    set_info_chord(data->d);
    goto end_switch;

case_insert_silence: {
    /* insert a new note at the current pos (-1 if we are at the end,
     * see insert_chord)
     */
    history_next_is_major(&data->d->history);
    clear_cursor(&data->d->cursor);
    if (!insert_chord(data->d, data->d->cursor.line,
          /* are we at the end of the line? */
          data->d->cursor.chord ==
            data->d->data[data->d->cursor.line].u.music.size ?
              -1 : data->d->cursor.chord, NOTES_CHORD)) {
      history_up_and_forget(data->d);
      size_post_treatment(data->d);
      CALL(error_state, data->g->win);
      goto end_switch;
    }
    if (!insert_note(data->d, data->d->cursor.line, data->d->cursor.chord,
                     SILENCE, 0, get_gui_duration(data->g), 0)) {
      history_up_and_forget(data->d);
      size_post_treatment(data->d);
      CALL(error_state, data->g->win);
      goto end_switch;
    }
    size_post_treatment(data->d);
    set_info_chord(data->d);
    goto end_switch;
  }

case_insert_note: {
    int n;
    /* set the note to insert */
    switch (key) {
      case 'a': n = 1; break;
      case 'b': n = 2; break;
      case 'c': n = 3; break;
      case 'd': n = 4; break;
      case 'e': n = 5; break;
      case 'f': n = -1; break;
      case 'g': n = 0; break;
    }
    n += get_delta_clef(data->d, data->d->cursor.line,
                        data->d->cursor.chord);
    /* insert a new note at the current pos (-1 if we are at the end,
     * see insert_chord)
     */
    history_next_is_major(&data->d->history);
    clear_cursor(&data->d->cursor);
    if (!insert_chord(data->d, data->d->cursor.line,
          /* are we at the end of the line? */
          data->d->cursor.chord ==
            data->d->data[data->d->cursor.line].u.music.size ?
              -1 : data->d->cursor.chord, NOTES_CHORD)) {
      history_up_and_forget(data->d);
      size_post_treatment(data->d);
      CALL(error_state, data->g->win);
      goto end_switch;
    }
    if (!insert_note(data->d, data->d->cursor.line, data->d->cursor.chord,
                     NOTE, n, NORMAL, get_gui_duration(data->g),
                     0, 0, 0, 0, 0)) {
      history_up_and_forget(data->d);
      size_post_treatment(data->d);
      CALL(error_state, data->g->win);
      goto end_switch;
    }
    size_post_treatment(data->d);
    set_info_chord(data->d);
    goto end_switch;
  }

case_tonalite: {
    /* insert a new note at the current pos (-1 if we are at the end,
     * see insert_chord)
     */
    history_next_is_major(&data->d->history);
    clear_cursor(&data->d->cursor);
    if (!insert_chord(data->d, data->d->cursor.line,
          /* are we at the end of the line? */
          data->d->cursor.chord ==
            data->d->data[data->d->cursor.line].u.music.size ?
              -1 : data->d->cursor.chord, TONALITE)) {
      history_up_and_forget(data->d);
      size_post_treatment(data->d);
      CALL(error_state, data->g->win);
      goto end_switch;
    }
    size_post_treatment(data->d);
    set_info_chord(data->d);
    goto end_switch;
  }

case_insert_break:
    history_next_is_major(&data->d->history);
    clear_cursor(&data->d->cursor);
    if (!insert_chord(data->d, data->d->cursor.line,
          /* are we at the end of the line? */
          data->d->cursor.chord ==
            data->d->data[data->d->cursor.line].u.music.size ?
              -1 : data->d->cursor.chord, MESURE_BREAK)) {
      history_up_and_forget(data->d);
      size_post_treatment(data->d);
      CALL(error_state, data->g->win);
      goto end_switch;
    }
    size_post_treatment(data->d);
    set_info_chord(data->d);
    goto end_switch;

case_insert_clef:
    history_next_is_major(&data->d->history);
    clear_cursor(&data->d->cursor);
    if (!insert_chord(data->d, data->d->cursor.line,
          /* are we at the end of the line? */
          data->d->cursor.chord ==
            data->d->data[data->d->cursor.line].u.music.size ?
              -1 : data->d->cursor.chord, CLEF)) {
      history_up_and_forget(data->d);
      size_post_treatment(data->d);
      CALL(error_state, data->g->win);
      goto end_switch;
    }
    size_post_treatment(data->d);
    set_info_chord(data->d);
    goto end_switch;

case_i: {
    /* no i-mode at end of line */
    if (data->d->cursor.chord ==
          data->d->data[data->d->cursor.line].u.music.size)
      goto end_switch;
    /* i-mode only for NOTES_CHORD */
    if (data->d->data[data->d->cursor.line].u.music.
          data[data->d->cursor.chord].type != NOTES_CHORD)
      goto end_switch;
    dprintf("call i state\n");
    CALL(i_state, data->d, data->g, &data->ret);
    dprintf("return from i state\n");
    if (data->ret == I_QUIT) RETURN;
    goto end_switch;
  }

case_play:
    /* we should call start_play in the enter stuff of play_state
     * but we can't use CaLL in enter stuff (yeah, bad hacking)
     * so this if has to be here
     */
    setup_tempo(data->d);
    if (!start_play(data->d, data->d->cursor.line, data->d->cursor.chord,
                    data->d->tempo)) {
      CALL(error_state, data->g->win);
      goto end_switch;
    }
    CALL(play_state, data->d);
    set_info_chord(data->d);
    goto end_switch;

case_goto_selection:
    CALL(state_selection, data->d, data->g, key, &data->ret);
    if (data->ret == I_QUIT) RETURN;
    goto end_switch;

case_paste_selection:
    history_next_is_major(&data->d->history);
    clear_cursor(&data->d->cursor);
    if (!paste_selection(data->d)) {
      history_up_and_forget(data->d);
      size_post_treatment(data->d);
      CALL(error_state, data->g->win);
      goto end_switch;
    }
    size_post_treatment(data->d);
    goto end_switch;

case_insert_text:
    /* beginning of line */
    if (data->d->cursor.chord == 0) {
      history_next_is_major(&data->d->history);
      clear_cursor(&data->d->cursor);
      if (!insert_text_line(data->d, data->d->cursor.line-1)) {
        history_up_and_forget(data->d);
        size_post_treatment(data->d);
        CALL(error_state, data->g->win);
        goto end_switch;
      }
      data->d->cursor.line++;
      size_post_treatment(data->d);
      goto end_switch;
    }
    /* end of line */
    else if (data->d->cursor.chord ==
               LINE_MUSIC_SIZE(data->d, data->d->cursor.line)) {
      history_next_is_major(&data->d->history);
      clear_cursor(&data->d->cursor);
      if (!insert_text_line(data->d, data->d->cursor.line)) {
        history_up_and_forget(data->d);
        size_post_treatment(data->d);
        CALL(error_state, data->g->win);
        goto end_switch;
      }
      data->d->cursor.line++;
      data->d->cursor.chord = 0;
      size_post_treatment(data->d);
      goto end_switch;
    }

end_switch:
    reset_cursor(&data->d->cursor);
    DO_GOTO_STATE(_owner);
  }

%%import "states/common_actions.i"

  action goto_state {
    dprintf("goto state\n");

    /* set the grey for undo/redo menu */
    set_grey_undo_redo(data->d);

    /* is it a text line? */
    if (LINE_TYPE(data->d, data->d->cursor.line) == LINE_TEXT) {
      CALL(text_state, data->d, data->g, &data->ret);
      if (data->ret == I_QUIT) RETURN;
      DO_GOTO_STATE(_owner);
    }
    /* if end of line, finished */
    else if (data->d->cursor.chord ==
               data->d->data[data->d->cursor.line].u.music.size)
      BREAK;
    /* if not, call appropriate state */
    else if (CHORD_TYPE(data->d, data->d->cursor.line,
                   data->d->cursor.chord) == TONALITE) {
      CALL(t_state, data->d, data->g, &data->key, &data->control,
           &data->shift, &data->ret);
      if (data->ret == I_QUIT) RETURN;
      if (data->ret == I_UNSUPPORTED_KEY)
        _DO_KEY_PRESS(_owner, data->key, data->control, data->shift);
      else
        DO_GOTO_STATE(_owner);
    } else if (CHORD_TYPE(data->d, data->d->cursor.line,
                   data->d->cursor.chord) == CLEF) {
      CALL(clef_state, data->d, data->g, &data->key, &data->control,
           &data->shift, &data->ret);
      if (data->ret == I_QUIT) RETURN;
      if (data->ret == I_UNSUPPORTED_KEY)
        _DO_KEY_PRESS(_owner, data->key, data->control, data->shift);
      else
        DO_GOTO_STATE(_owner);
    } else if (CHORD_TYPE(data->d, data->d->cursor.line,
                          data->d->cursor.chord) == NOTES_CHORD &&
               CHORD_SIZE(data->d, data->d->cursor.line,
                          data->d->cursor.chord) == 1) {
      CALL(m_state, data->d, data->g, &data->key, &data->control,
           &data->shift, &data->ret);
      if (data->ret == I_QUIT) RETURN;
      if (data->ret == I_UNSUPPORTED_KEY)
        _DO_KEY_PRESS(_owner, data->key, data->control, data->shift);
      else
        DO_GOTO_STATE(_owner);
    }
  }
}
