/* gcomposer: a little tool to write music.
 * http://sed.free.fr/gcomposer
 * Released in the public domain (freedom has no license).
 */

state m_state(doc *d, gui *g, int *key, int *control, int *shift,
              int *ret_code)
{
  data {
    doc *d;
    struct gui *g;
    int *key;
    int *control;
    int *shift;
    int *ret_code;
    char load_file[FILENAME_MAX];
  }

  enter {
    dprintf("enter state m cursor = %d\n", d->cursor.type);
    data->d = d;
    data->g = g;
    data->ret_code = ret_code;
    data->key = key;
    data->control = control;
    data->shift = shift;
    *data->ret_code = I_NOTHING;
  }

  leave {
    dprintf("leave state m cursor = %d\n", data->d->cursor.type);
    /* do nothing */
  }

  action key_press(int key, int control, int shift) {
    int pitch_direction;

    if (control) {
      switch (key) {
        case 'l': pitch_direction = -1; goto case_head_delta;
        case 'L': pitch_direction = 1; goto case_head_delta;
        case 'm': pitch_direction = -1; goto case_alteration_delta;
        case 'M': pitch_direction = 1; goto case_alteration_delta;
        case 'o': pitch_direction = -1; goto case_dots_delta;
        case 'O': pitch_direction = 1; goto case_dots_delta;
      }
      goto not_handled_key;
    }

    switch (key) {
      default: goto not_handled_key;
      case '#': goto case_diese;
      case '.': goto case_point;
      case 't': pitch_direction = 1; goto case_time;
      case 'T': pitch_direction = -1; goto case_time;
      case 'u': pitch_direction = 1; goto case_pitch;
      case 'U': pitch_direction = -1; goto case_pitch;
      case '0': case '1': case '2': case '3': case '4':
      case '5': case '6': case '7': case '8': case '9':
        goto case_octave;
      case 'l': pitch_direction = -1; goto case_head_delta;
      case 'L': pitch_direction = 1; goto case_head_delta;
      case 'm': pitch_direction = -1; goto case_alteration_delta;
      case 'M': pitch_direction = 1; goto case_alteration_delta;
      case 'o': pitch_direction = -1; goto case_dots_delta;
      case 'O': pitch_direction = 1; goto case_dots_delta;
      case 'j': pitch_direction = -1; goto case_stem;
      case 'J': pitch_direction = 1; goto case_stem;
    }

case_pitch: {
    int note_number = CHORD_SIZE(data->d, data->d->cursor.line,
                                 data->d->cursor.chord)-1;

    switch (NOTE_TYPE(data->d, data->d->cursor.line,
                      data->d->cursor.chord, note_number)) {
      case NOTE: {
        int pitch = NOTE_PITCH(data->d, data->d->cursor.line,
                               data->d->cursor.chord, note_number);
        pitch += pitch_direction;
        history_next_is_major(&data->d->history);
        clear_cursor(&data->d->cursor);
        if (set_note_pitch(data->d, data->d->cursor.line,
                           data->d->cursor.chord,
                           note_number, pitch) < 0) {
          history_up_and_forget(data->d);
          size_post_treatment(data->d);
          goto set_error;
        }
        size_post_treatment(data->d);
        break;
      }
      case SILENCE: {
        int position = SILENCE_POSITION(data->d, data->d->cursor.line,
                                        data->d->cursor.chord, note_number);
        position += pitch_direction;
        history_next_is_major(&data->d->history);
        clear_cursor(&data->d->cursor);
        if (set_note_position(data->d, data->d->cursor.line,
                              data->d->cursor.chord,
                              note_number, position) < 0) {
          history_up_and_forget(data->d);
          size_post_treatment(data->d);
          goto set_error;
        }
        size_post_treatment(data->d);
        break;
      }
    }
    set_info_chord(data->d);
    goto end_switch;
  }

case_diese: {
    int note_number = CHORD_SIZE(data->d, data->d->cursor.line,
                                 data->d->cursor.chord)-1;
    int alteration;

    if (NOTE_TYPE(data->d, data->d->cursor.line,
                  data->d->cursor.chord, note_number) != NOTE)
      goto end_switch;
    alteration = NOTE_ALTERATION(data->d, data->d->cursor.line,
                                 data->d->cursor.chord, note_number);
    switch (alteration) {
      case NORMAL: alteration = DIESE; break;
      case DIESE: alteration = BEMOL; break;
      case BEMOL: alteration = BECARRE; break;
      case BECARRE: alteration = NORMAL; break;
    }
    history_next_is_major(&data->d->history);
    clear_cursor(&data->d->cursor);
    if (set_note_alteration(data->d, data->d->cursor.line,
                            data->d->cursor.chord,
                            note_number, alteration) < 0) {
      history_up_and_forget(data->d);
      size_post_treatment(data->d);
      goto set_error;
    }
    size_post_treatment(data->d);
    set_info_chord(data->d);
    goto end_switch;
  }

case_point: {
    int note_number = CHORD_SIZE(data->d, data->d->cursor.line,
                                 data->d->cursor.chord)-1;
    int nb_dots;

    switch (NOTE_TYPE(data->d, data->d->cursor.line,
                      data->d->cursor.chord, note_number)) {
      case NOTE:
        nb_dots = NOTE_NB_DOTS(data->d, data->d->cursor.line,
                               data->d->cursor.chord, note_number);
        break;
      case SILENCE:
        nb_dots = SILENCE_NB_DOTS(data->d, data->d->cursor.line,
                                  data->d->cursor.chord, note_number);
        break;
    }
    nb_dots++;
    if (nb_dots == 4) nb_dots = 0;
    history_next_is_major(&data->d->history);
    clear_cursor(&data->d->cursor);
    if (set_note_dot(data->d, data->d->cursor.line,
                     data->d->cursor.chord,
                     note_number, nb_dots) < 0) {
      history_up_and_forget(data->d);
      size_post_treatment(data->d);
      goto set_error;
    }
    size_post_treatment(data->d);
    set_info_chord(data->d);
    goto end_switch;
  }

case_time: {
    int note_number = CHORD_SIZE(data->d, data->d->cursor.line,
                                 data->d->cursor.chord)-1;
    int duration;

    switch (NOTE_TYPE(data->d, data->d->cursor.line,
                      data->d->cursor.chord, note_number)) {
      case NOTE:
        duration = NOTE_DURATION(data->d, data->d->cursor.line,
                                 data->d->cursor.chord, note_number);
        break;
      case SILENCE:
        duration = SILENCE_DURATION(data->d, data->d->cursor.line,
                                    data->d->cursor.chord, note_number);
        break;
    }
    switch (duration) {
      case RONDE:
        if (pitch_direction == 1) duration = BLANCHE;
        break;
      case BLANCHE:
        if (pitch_direction == 1) duration = NOIRE; else duration = RONDE;
        break;
      case NOIRE:
        if (pitch_direction == 1) duration = CROCHE; else duration = BLANCHE;
        break;
      case CROCHE:
        if (pitch_direction == 1) duration = DCROCHE; else duration = NOIRE;
        break;
      case DCROCHE:
        if (pitch_direction == 1) duration = DDCROCHE; else duration = CROCHE;
        break;
      case DDCROCHE:
        if (pitch_direction == 1) duration = DDDCROCHE;
        else duration = DCROCHE;
        break;
      case DDDCROCHE:
        if (pitch_direction == 1) duration = DDDDCROCHE;
        else duration = DDCROCHE;
        break;
      case DDDDCROCHE:
        if (pitch_direction == -1) duration = DDDCROCHE;
        break;
    }
    history_next_is_major(&data->d->history);
    clear_cursor(&data->d->cursor);
    if (set_note_duration(data->d, data->d->cursor.line,
                          data->d->cursor.chord,
                          note_number, duration) < 0) {
      history_up_and_forget(data->d);
      size_post_treatment(data->d);
      goto set_error;
    }
    size_post_treatment(data->d);
    set_info_chord(data->d);
    goto end_switch;
  }

case_octave: {
    int octave = key - '0';
    int note_number = CHORD_SIZE(data->d, data->d->cursor.line,
                                 data->d->cursor.chord)-1;

    switch (NOTE_TYPE(data->d, data->d->cursor.line,
                      data->d->cursor.chord, note_number)) {
      case NOTE: {
        int pitch = NOTE_PITCH(data->d, data->d->cursor.line,
                               data->d->cursor.chord, note_number);
        int delta_clef = get_delta_clef(data->d, data->d->cursor.line,
                                        data->d->cursor.chord);
        pitch -= delta_clef;
        /* redo, too slow (but hey, it won't loop 1,000,000 times...) */
        /* the 4th octave is from C4 to B4, thus the -4 and 3 below */
        while (pitch < -4) pitch += 7;
        while (pitch >= 3) pitch -= 7;
        pitch += 7 * (octave - 4);
        pitch += delta_clef;
        /* pitch not changed? */
        if (pitch == NOTE_PITCH(data->d, data->d->cursor.line,
                                data->d->cursor.chord, note_number))
          goto end_switch;
        history_next_is_major(&data->d->history);
        clear_cursor(&data->d->cursor);
        if (set_note_pitch(data->d, data->d->cursor.line,
                           data->d->cursor.chord,
                           note_number, pitch) < 0) {
          history_up_and_forget(data->d);
          size_post_treatment(data->d);
          goto set_error;
        }
        size_post_treatment(data->d);
        break;
      }
      case SILENCE: /* do nothing */; break;
    }
    set_info_chord(data->d);
    goto end_switch;
  }

case_head_delta: {
    int note_number = CHORD_SIZE(data->d, data->d->cursor.line,
                                 data->d->cursor.chord)-1;
    if (NOTE_TYPE(data->d, data->d->cursor.line,
                  data->d->cursor.chord, note_number) != NOTE)
      goto end_switch;
    if (!control)
      pitch_direction *= images[I_BLACK_HEAD].width-1;
    pitch_direction += NOTE_HEAD_DELTA(data->d,
                                       data->d->cursor.line,
                                       data->d->cursor.chord,
                                       note_number);
    history_next_is_major(&data->d->history);
    clear_cursor(&data->d->cursor);
    if (set_note_head_delta(data->d, data->d->cursor.line,
                            data->d->cursor.chord,
                            note_number, pitch_direction) < 0) {
      history_up_and_forget(data->d);
      size_post_treatment(data->d);
      goto set_error;
    }
    size_post_treatment(data->d);
    set_info_chord(data->d);
    goto end_switch;
  }

case_alteration_delta: {
    int note_number = CHORD_SIZE(data->d, data->d->cursor.line,
                                 data->d->cursor.chord)-1;
    if (NOTE_TYPE(data->d, data->d->cursor.line,
                  data->d->cursor.chord, note_number) != NOTE)
      goto end_switch;
    if (!control)
      pitch_direction *= images[I_BLACK_HEAD].width-1;
    pitch_direction += NOTE_ALTERATION_DELTA(data->d,
                                             data->d->cursor.line,
                                             data->d->cursor.chord,
                                             note_number);
    history_next_is_major(&data->d->history);
    clear_cursor(&data->d->cursor);
    if (set_note_alteration_delta(data->d, data->d->cursor.line,
                                  data->d->cursor.chord,
                                  note_number, pitch_direction) < 0) {
      history_up_and_forget(data->d);
      size_post_treatment(data->d);
      goto set_error;
    }
    size_post_treatment(data->d);
    set_info_chord(data->d);
    goto end_switch;
  }

case_dots_delta: {
    int note_number = CHORD_SIZE(data->d, data->d->cursor.line,
                                 data->d->cursor.chord)-1;
    if (NOTE_TYPE(data->d, data->d->cursor.line,
                  data->d->cursor.chord, note_number) != NOTE)
      goto end_switch;
    if (!control)
      pitch_direction *= images[I_BLACK_HEAD].width-1;
    pitch_direction += NOTE_DOTS_DELTA(data->d,
                                       data->d->cursor.line,
                                       data->d->cursor.chord,
                                       note_number);
    history_next_is_major(&data->d->history);
    clear_cursor(&data->d->cursor);
    if (set_note_dots_delta(data->d, data->d->cursor.line,
                            data->d->cursor.chord,
                            note_number, pitch_direction) < 0) {
      history_up_and_forget(data->d);
      size_post_treatment(data->d);
      goto set_error;
    }
    size_post_treatment(data->d);
    set_info_chord(data->d);
    goto end_switch;
  }

case_stem: {
    int note_number = CHORD_SIZE(data->d, data->d->cursor.line,
                                 data->d->cursor.chord)-1;
    if (NOTE_TYPE(data->d, data->d->cursor.line,
                  data->d->cursor.chord, note_number) != NOTE)
      goto end_switch;
    history_next_is_major(&data->d->history);
    clear_cursor(&data->d->cursor);
    if (set_note_stem_down(data->d, data->d->cursor.line,
                           data->d->cursor.chord,
                           note_number, pitch_direction == -1) < 0) {
      history_up_and_forget(data->d);
      size_post_treatment(data->d);
      goto set_error;
    }
    size_post_treatment(data->d);
    set_info_chord(data->d);
    goto end_switch;
  }

not_handled_key:
    *data->key = key;
    *data->control = control;
    *data->shift = shift;
    *data->ret_code = I_UNSUPPORTED_KEY;
    RETURN;

set_error:
    CALL(error_state, data->g->win);
    goto end_switch;

end_switch:
    reset_cursor(&data->d->cursor);

    /* set the grey for undo/redo */
    set_grey_undo_redo(data->d);
  }

%%import "states/common_actions.i"
}
