/* public domain stuff */

#include "symlist.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "sstring.h"

typedef struct {
  unsigned long addr;
  unsigned long size;
  int name;
} symbol;

/* globals are The Truth */
sym_info *symbols;
int *sorted_symbols_by_address;
int *sorted_symbols_by_name;
int *sorted_symbols_by_calls;
int *sorted_symbols_by_self;
int *sorted_symbols_by_cumuled_1;
int *sorted_symbols_by_cumuled_2;

int *sorted_data;

/* local globals are The Half-Truth */
static symbol *data;
static int size;
static int unique_size;
static int maxsize;

int overlap(unsigned long min1, unsigned long max1,
            unsigned long min2, unsigned long max2)
{
  unsigned long min, max;
  if (max1 < min2) return -1;
  if (min1 > max2) return -2;
  if (min1 <= min2) min = min2; else min = min1;
  if (max1 >= max2) max = max2; else max = max1;
  return max-min+1;
}   

unsigned long get_symbol_lowpc(int i)
{
  return data[i].addr;
}

unsigned long get_symbol_highpc(int i)
{
  return data[i].addr+data[i].size-1;
}

int nb_symbols(void)
{
  return size;
}

int nb_unique_symbols(void)
{
  return unique_size;
}

int add_null_symbol(void)
{
  if (size == maxsize) {
    symbol *new;
    int ns = maxsize + 2048;
    if (data)
      new = realloc(data, ns * sizeof(symbol));
    else
      new = malloc(ns * sizeof(symbol));
    if (!new) return -1;
    data = new;
    maxsize = ns;
  }
  data[size].addr = 0;
  data[size].size = 1; /*dummy size of 1 (0 leads to bug in histogram stuff)*/
  data[size].name = new_string("<not found>");
  if (data[size].name < 0) return -1;
  size++;
  return size-1;
}

int add_symbol(unsigned long addr, unsigned long s, char *name)
{
  if (addr != 0 && s == 0)
    fprintf(stderr, "WARNING null sized symbol %s (%8.8lx)\n", name, addr);

  if (addr == 0 || s == 0) return 0;

  if (size == maxsize) {
    symbol *new;
    int ns = maxsize + 2048;
    if (data)
      new = realloc(data, ns * sizeof(symbol));
    else
      new = malloc(ns * sizeof(symbol));
    if (!new) return -1;
    data = new;
    maxsize = ns;
  }
  data[size].addr = addr;
  data[size].size = s;
  data[size].name = new_string(name);
  if (data[size].name < 0) return -1;
  size++;
  return size-1;
}

static int qsort_cmp(const void *a, const void *b)
{
  int aa = *(int*)a;
  int bb = *(int*)b;
  symbol *sa = data+aa;
  symbol *sb = data+bb;
  if (sa->addr < sb->addr) return -1;
  if (sa->addr > sb->addr) return 1;
  return 0;
}

void dump_symbol(symbol *s)
{
  fprintf(stdout, "%8.8lx %lu %s\n", s->addr, s->size, get_string(s->name));
}

extern int overlap(unsigned long, unsigned long,
                   unsigned long, unsigned long);

static int uniquify(void)
{
  /* don't ask me why, symbols may appear more than once
   * so we do it the barbarian way: if symbol A and B share
   * same starting address, remove B
   */
//  int *resorted_data = malloc(size * sizeof(int));
//  if (!resorted_data) return -1;
  int i;
  int n = 0;
  symbol *sym, *cur;
//  /* useless memcpy, should swap pointers */
//  memcpy(resorted_data, sorted_data, size * sizeof(int));
  for (i = 1; i<size; i++) {
//    sym = data+resorted_data[i];
    sym = data+sorted_data[i];
    cur = data+sorted_data[n];
    if (sym->addr == cur->addr) {
      fprintf(stderr,
"WARNING symbol %s same start address than %s (%8.8lx %8.8lx %8.8lx %8.8lx)",
              get_string(sym->name), get_string(cur->name),
              sym->addr, sym->size, cur->addr, cur->size);
      fprintf(stderr, " (symbol %s is removed)\n", get_string(sym->name));
      continue;
    }
    n++;
//    sorted_data[n] = resorted_data[i];
    if (i != n) sorted_data[n] = sorted_data[i];
  }
  unique_size = n+1;
//  free(resorted_data);
  return 0;
}

int end_add_symbol(void)
{
  int i;
  sorted_data = malloc(size * sizeof(int));
  if (!sorted_data) return -1;
  for (i = 0; i < size; i++) sorted_data[i] = i;
  qsort(sorted_data, size, sizeof(int), qsort_cmp);

  if (uniquify() == -1) return -1;

  for (i=0; i<unique_size-1; i++) {
    symbol *a, *b;
    a = data+sorted_data[i];
    b = data+sorted_data[i+1];
    unsigned long first_low = a->addr;
    unsigned long first_high = a->addr + a->size - 1;
    unsigned long second_low = b->addr;
    unsigned long second_high = b->addr + b->size - 1;
    if (overlap(first_low, first_high, second_low, second_high) >= 0) {
      fprintf(stderr,
 "WARNING symbol %s overlaps the next one %s (%8.8lx %8.8lx %8.8lx %8.8lx)\n",
              get_string(a->name),
              get_string(b->name),
              first_low, first_high, second_low, second_high);
      if (first_low >= second_low)
        fprintf(stderr,
          "  cannot change end address, second is less or egal than first\n");
      else {
        fprintf(stderr,
          "  changing end address of 1st symbol so no more overlap\n");
        a->size = second_low-first_low;
      }
    }
  }

  return 0;
}

static int
__attribute__((__no_instrument_function__))
cmp(symbol *b, unsigned long dig_addr)
{
  if (b->addr <= dig_addr) {
    if (b->addr + b->size > dig_addr) return 0;
    return -1;
  }
  return 1;
}

int
__attribute__((__no_instrument_function__))
find_symbol(unsigned long dig_addr)
{
  int min = 0;
  int max = unique_size-1;
  int cur;
  int tst;
  symbol *sym;

  while (1) {
    if (max < min) return -1;
    cur = (max+min)/2;
    sym = data+sorted_data[cur];
    tst = cmp(sym, dig_addr);
    if (tst > 0)
      max = cur-1;
    else if (tst < 0)
      min = cur+1;
    else
      return sorted_data[cur];
  }

  /* dummy return */
  return -1;
}

char *
__attribute__((__no_instrument_function__))
symbol_name(int symbol)
{
  return get_string(data[symbol].name);
}

static int count_sort(const void *a, const void *b)
{
  int aa = *(int*)a;
  int bb = *(int*)b;
  sym_info *sa = symbols + aa;
  sym_info *sb = symbols + bb;
  if (sa->count < sb->count) return -1;
  if (sa->count > sb->count) return 1;
  return 0;
}

static int count_sort_by_self(const void *a, const void *b)
{
  int aa = *(int*)a;
  int bb = *(int*)b;
  sym_info *sa = symbols + aa;
  sym_info *sb = symbols + bb;
  if (sa->internal_time < sb->internal_time) return -1;
  if (sa->internal_time > sb->internal_time) return 1;
  /* same self, order by count */
  if (sa->count < sb->count) return -1;
  if (sa->count > sb->count) return 1;
  /* same count, order as qsort wants */
  return 0;
} 

static int count_sort_by_cumuled_1(const void *a, const void *b)
{
  int aa = *(int*)a;
  int bb = *(int*)b;
  sym_info *sa = symbols + aa;
  sym_info *sb = symbols + bb;
  if (sa->cumuled_time_1 < sb->cumuled_time_1) return -1;
  if (sa->cumuled_time_1 > sb->cumuled_time_1) return 1;
  /* same cumuled time, order by self time */
  if (sa->internal_time < sb->internal_time) return -1;
  if (sa->internal_time > sb->internal_time) return 1;
  /* same self, order by count */
  if (sa->count < sb->count) return -1;
  if (sa->count > sb->count) return 1;
  /* same count, order as qsort wants */
  return 0;
}

static int count_sort_by_cumuled_2(const void *a, const void *b)
{
  int aa = *(int*)a;
  int bb = *(int*)b;
  sym_info *sa = symbols + aa;
  sym_info *sb = symbols + bb;
  if (sa->cumuled_time_2 < sb->cumuled_time_2) return -1;
  if (sa->cumuled_time_2 > sb->cumuled_time_2) return 1;
  /* same cumuled time, order by self time */
  if (sa->internal_time < sb->internal_time) return -1;
  if (sa->internal_time > sb->internal_time) return 1;
  /* same self, order by count */
  if (sa->count < sb->count) return -1;
  if (sa->count > sb->count) return 1;
  /* same count, order as qsort wants */
  return 0;
}

static int count_sort_by_name(const void *a, const void *b)
{
  int aa = *(int*)a;
  int bb = *(int*)b;
  return strcmp(symbol_name(aa), symbol_name(bb));
}

void uniquify_again(void)
{
  int i;
  int n = 0;
  sym_info *sym;
  for (i = 0; i<size; i++) {
    sym = symbols+sorted_data[i];
    if (sym->count == 0 && sym->internal_time == 0 &&
        list_size(&sym->call) == 0) {
#if 0
      fprintf(stderr, "WARNING symbol %s is not used (removed)\n",
              symbol_name(sorted_data[i]));
#endif
      continue;
    }
    if (i != n) sorted_data[n] = sorted_data[i];
    n++;
  }
  unique_size = n;
}

int sort_symbols(void)
{
  int n;
  int i;

  uniquify_again();

  n = nb_unique_symbols();

  /* by name */
  sorted_symbols_by_name = malloc(n * sizeof(int));
  if (!sorted_symbols_by_name) return -1;
  for (i = 0; i < n; i++) sorted_symbols_by_name[i] = sorted_data[i];
  qsort(sorted_symbols_by_name, n, sizeof(int), count_sort_by_name);

  /* by # calls */
  sorted_symbols_by_calls = malloc(n * sizeof(int));
  if (!sorted_symbols_by_calls) return -1;
  for (i = 0; i < n; i++) sorted_symbols_by_calls[i] = sorted_data[i];
  qsort(sorted_symbols_by_calls, n, sizeof(int), count_sort);

  /* by self */
  sorted_symbols_by_self = malloc(n * sizeof(int));
  if (!sorted_symbols_by_self) return -1;
  for (i = 0; i < n; i++) sorted_symbols_by_self[i] = sorted_data[i];
  qsort(sorted_symbols_by_self, n, sizeof(int), count_sort_by_self);

  /* by cumul 1 */
  sorted_symbols_by_cumuled_1 = malloc(n * sizeof(int));
  if (!sorted_symbols_by_cumuled_1) return -1;
  for (i = 0; i < n; i++) sorted_symbols_by_cumuled_1[i] = sorted_data[i];
  qsort(sorted_symbols_by_cumuled_1, n, sizeof(int), count_sort_by_cumuled_1);

  /* by cumul 2 */
  sorted_symbols_by_cumuled_2 = malloc(n * sizeof(int));
  if (!sorted_symbols_by_cumuled_2) return -1;
  for (i = 0; i < n; i++) sorted_symbols_by_cumuled_2[i] = sorted_data[i];
  qsort(sorted_symbols_by_cumuled_2, n, sizeof(int), count_sort_by_cumuled_2);

  return 0;
}
